import mock
import pytest

from uaclient import apt, exceptions, messages, util
from uaclient.entitlements.entitlement_status import (
    ApplicabilityStatus,
    ApplicationStatus,
)
from uaclient.entitlements.repo import RepoEntitlement
from uaclient.testing.helpers import does_not_raise

M_PATH = "uaclient.entitlements.repo."
M_CONTRACT_PATH = "uaclient.entitlements.repo.contract.UAContractClient."


class RepoTestEntitlement(RepoEntitlement):
    """Subclass so we can test shared repo functionality"""

    name = "repotest"
    title = "Repo Test Class"
    description = "Repo entitlement for testing"
    repo_key_file = "test.gpg"


class RepoTestEntitlementRepoWithPin(RepoTestEntitlement):
    repo_pin_priority = 1000


class RepoTestEntitlementRepoWithRemovePackages(RepoTestEntitlement):
    def remove_packages(self):
        pass


@pytest.fixture
def entitlement(entitlement_factory):
    return entitlement_factory(
        RepoTestEntitlement, affordances={"series": ["xenial"]}
    )


class TestProcessContractDeltas:
    @pytest.mark.parametrize("orig_access", ({}, {"entitlement": {}}))
    def test_on_no_deltas(self, orig_access):
        """Return True when no deltas are available to process."""
        entitlement = RepoTestEntitlement()
        with mock.patch.object(
            entitlement, "remove_apt_config"
        ) as m_remove_apt_config:
            with mock.patch.object(
                entitlement, "setup_apt_config"
            ) as m_setup_apt_config:
                assert entitlement.process_contract_deltas(orig_access, {})
        assert [] == m_remove_apt_config.call_args_list
        assert [] == m_setup_apt_config.call_args_list

    @pytest.mark.parametrize("entitled", (False, util.DROPPED_KEY))
    @mock.patch.object(RepoTestEntitlement, "disable")
    @mock.patch.object(
        RepoTestEntitlement, "can_disable", return_value=(True, None)
    )
    @mock.patch.object(RepoTestEntitlement, "application_status")
    def test_disable_when_delta_to_unentitled(
        self,
        m_application_status,
        m_can_disable,
        m_disable,
        entitlement,
        entitled,
    ):
        """Disable the service on contract transitions to unentitled."""
        application_status = ApplicationStatus.ENABLED
        m_application_status.return_value = (application_status, "")
        assert entitlement.process_contract_deltas(
            {"entitlement": {"entitled": True}},
            {"entitlement": {"entitled": entitled}},
        )
        assert [mock.call(mock.ANY)] == m_disable.call_args_list

    @mock.patch.object(RepoTestEntitlement, "remove_apt_config")
    @mock.patch.object(RepoTestEntitlement, "application_status")
    @mock.patch.object(RepoTestEntitlement, "applicability_status")
    def test_no_changes_when_service_inactive_and_not_enable_by_default(
        self,
        m_applicability_status,
        m_application_status,
        m_remove_apt_config,
        entitlement,
    ):
        """Noop when service is inactive and not enableByDefault."""
        m_application_status.return_value = (
            ApplicationStatus.DISABLED,
            "",
        )
        m_applicability_status.return_value = (
            ApplicabilityStatus.APPLICABLE,
            "",
        )
        assert not entitlement.process_contract_deltas(
            {"entitlement": {"entitled": True}},
            {
                "entitlement": {"obligations": {"enableByDefault": False}},
                "resourceToken": "repotest-token",
            },
        )
        assert [] == m_remove_apt_config.call_args_list

    @mock.patch.object(RepoTestEntitlement, "enable")
    @mock.patch.object(RepoTestEntitlement, "application_status")
    @mock.patch.object(RepoTestEntitlement, "applicability_status")
    def test_allow_enable_when_inactive_enable_by_default_and_resource_token(
        self,
        m_applicability_status,
        m_application_status,
        m_enable,
        entitlement,
    ):
        """Update apt when inactive, enableByDefault and allow_enable."""
        m_application_status.return_value = (
            ApplicationStatus.DISABLED,
            "",
        )
        m_applicability_status.return_value = (
            ApplicabilityStatus.APPLICABLE,
            "",
        )
        assert entitlement.process_contract_deltas(
            {"entitlement": {"entitled": True}},
            {
                "entitlement": {"obligations": {"enableByDefault": True}},
                "resourceToken": "repotest-token",
            },
            allow_enable=True,
        )
        assert [mock.call(mock.ANY)] == m_enable.call_args_list

    @mock.patch.object(RepoTestEntitlement, "enable")
    @mock.patch.object(RepoTestEntitlement, "application_status")
    @mock.patch.object(RepoTestEntitlement, "applicability_status")
    def test_not_allow_enable_logs_message_when_inactive_enable_by_default(
        self,
        m_applicability_status,
        m_application_status,
        m_enable,
        entitlement,
        capsys,
        event,
    ):
        """Log a message when inactive, enableByDefault and allow_enable."""
        m_application_status.return_value = (
            ApplicationStatus.DISABLED,
            "",
        )
        m_applicability_status.return_value = (
            ApplicabilityStatus.APPLICABLE,
            "",
        )
        assert entitlement.process_contract_deltas(
            {"entitlement": {"entitled": True}},
            {
                "entitlement": {"obligations": {"enableByDefault": True}},
                "resourceToken": "repotest-token",
            },
            allow_enable=False,
        )
        assert [] == m_enable.call_args_list
        expected_msg = messages.ENABLE_BY_DEFAULT_MANUAL_TMPL.format(
            name="repotest"
        )
        assert expected_msg in capsys.readouterr()[1]

    @pytest.mark.parametrize("packages", ([], ["extremetuxracer"]))
    @mock.patch(
        "uaclient.files.state_files.status_cache_file.read", return_value=None
    )
    @mock.patch.object(RepoTestEntitlement, "install_packages")
    @mock.patch(M_PATH + "apt.remove_auth_apt_repo")
    @mock.patch.object(RepoTestEntitlement, "setup_apt_config")
    @mock.patch.object(RepoTestEntitlement, "remove_apt_config")
    @mock.patch.object(RepoTestEntitlement, "application_status")
    @mock.patch.object(RepoTestEntitlement, "_check_apt_url_is_applied")
    def test_update_apt_config_and_install_packages_when_active(
        self,
        m_check_apt_url_applied,
        m_application_status,
        m_remove_apt_config,
        m_setup_apt_config,
        m_remove_auth_apt_repo,
        m_install_packages,
        _m_status_cache,
        packages,
        entitlement,
    ):
        """Update_apt_config and packages if active and not enableByDefault."""
        m_check_apt_url_applied.return_value = False
        application_status = ApplicationStatus.ENABLED
        m_application_status.return_value = (application_status, "")
        deltas = {
            "entitlement": {"obligations": {"enableByDefault": False}},
            "resourceToken": "repotest-token",
        }
        if packages:
            deltas["entitlement"] = {
                "directives": {"additionalPackages": packages}
            }

        assert entitlement.process_contract_deltas(
            {"entitlement": {"entitled": True}}, deltas
        )
        assert [mock.call(mock.ANY)] == m_remove_apt_config.call_args_list
        assert [mock.call(mock.ANY)] == m_setup_apt_config.call_args_list
        if packages:
            assert [
                mock.call(mock.ANY, package_list=packages)
            ] == m_install_packages.call_args_list
        else:
            assert 0 == m_install_packages.call_count
        assert [] == m_remove_auth_apt_repo.call_args_list
        assert 1 == m_check_apt_url_applied.call_count

    @pytest.mark.parametrize(
        "series,file_extension", (("jammy", "list"), ("noble", "sources"))
    )
    @mock.patch("uaclient.files.state_files.status_cache_file.read")
    @mock.patch(
        "uaclient.entitlements.base.UAEntitlement.process_contract_deltas"
    )
    @mock.patch(M_PATH + "system.get_release_info")
    @mock.patch(M_PATH + "apt.remove_auth_apt_repo")
    @mock.patch.object(RepoTestEntitlement, "setup_apt_config")
    @mock.patch.object(RepoTestEntitlement, "remove_apt_config")
    @mock.patch.object(RepoTestEntitlement, "_check_apt_url_is_applied")
    def test_remove_old_auth_apt_repo_when_active_and_apt_url_delta(
        self,
        m_check_apt_url_applied,
        m_remove_apt_config,
        m_setup_apt_config,
        m_remove_auth_apt_repo,
        m_release_info,
        m_process_contract_deltas,
        m_status_cache_read,
        series,
        file_extension,
        entitlement,
    ):
        """Remove old apt url when aptURL delta occurs on active service."""
        m_check_apt_url_applied.return_value = False
        m_process_contract_deltas.return_value = False
        m_status_cache_read.return_value = {
            "services": [{"name": "repotest", "status": "enabled"}]
        }
        m_release_info.return_value.series = series
        assert entitlement.process_contract_deltas(
            {
                "entitlement": {
                    "entitled": True,
                    "directives": {"aptURL": "http://old"},
                }
            },
            {
                "entitlement": {
                    "obligations": {"enableByDefault": False},
                    "directives": {"aptURL": "http://new"},
                },
                "resourceToken": "repotest-token",
            },
        )
        assert [mock.call(mock.ANY)] == m_remove_apt_config.call_args_list
        assert [mock.call(mock.ANY)] == m_setup_apt_config.call_args_list
        apt_auth_remove_calls = [
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-repotest.{}".format(
                    file_extension
                ),
                "http://old",
                [],
            )
        ]
        assert apt_auth_remove_calls == m_remove_auth_apt_repo.call_args_list
        assert [
            mock.call(),
            mock.call(),
        ] == m_status_cache_read.call_args_list
        assert 1 == m_process_contract_deltas.call_count

        assert [
            mock.call("http://new")
        ] == m_check_apt_url_applied.call_args_list
        assert 1 == m_check_apt_url_applied.call_count

    @mock.patch(
        "uaclient.entitlements.base.UAEntitlement.process_contract_deltas"
    )
    @mock.patch("uaclient.files.state_files.status_cache_file.read")
    @mock.patch(M_PATH + "system.get_release_info")
    @mock.patch(M_PATH + "apt.remove_auth_apt_repo")
    @mock.patch.object(RepoTestEntitlement, "setup_apt_config")
    @mock.patch.object(RepoTestEntitlement, "remove_apt_config")
    @mock.patch.object(RepoTestEntitlement, "_check_apt_url_is_applied")
    def test_system_does_not_change_when_apt_url_delta_already_applied(
        self,
        m_check_apt_url_applied,
        m_remove_apt_config,
        m_setup_apt_config,
        m_remove_auth_apt_repo,
        m_release_info,
        m_status_cache_read,
        m_process_contract_deltas,
        entitlement,
    ):
        """Do not change system if apt url delta is already applied."""
        m_check_apt_url_applied.return_value = True
        m_process_contract_deltas.return_value = False
        m_status_cache_read.return_value = {
            "services": [{"name": "repotest", "status": "enabled"}]
        }
        assert entitlement.process_contract_deltas(
            {
                "entitlement": {
                    "entitled": True,
                    "directives": {"aptURL": "http://old"},
                }
            },
            {
                "entitlement": {
                    "obligations": {"enableByDefault": False},
                    "directives": {"aptURL": "http://new"},
                },
                "resourceToken": "repotest-token",
            },
        )
        assert m_remove_apt_config.call_count == 0
        assert m_setup_apt_config.call_count == 0
        assert m_remove_auth_apt_repo.call_count == 0

        assert [
            mock.call(),
            mock.call(),
        ] == m_status_cache_read.call_args_list
        assert 1 == m_process_contract_deltas.call_count

        assert [
            mock.call("http://new")
        ] == m_check_apt_url_applied.call_args_list
        assert 1 == m_check_apt_url_applied.call_count


class TestRepoEnable:
    @pytest.mark.parametrize("with_pre_install_msg", (False, True))
    @pytest.mark.parametrize("packages", (["a"], [], None))
    @pytest.mark.parametrize(
        "series,file_extension", (("xenial", "list"), ("noble", "sources"))
    )
    @mock.patch("uaclient.apt.update_sources_list")
    @mock.patch("uaclient.apt.setup_apt_proxy")
    @mock.patch(M_PATH + "system.subp", return_value=("", ""))
    @mock.patch(M_PATH + "apt.add_auth_apt_repo")
    @mock.patch(M_PATH + "exists", return_value=True)
    @mock.patch(M_PATH + "system.get_release_info")
    @mock.patch.object(
        RepoTestEntitlement, "can_enable", return_value=(True, None)
    )
    def test_enable_calls_adds_apt_repo_and_calls_apt_update(
        self,
        m_can_enable,
        m_release_info,
        m_exists,
        m_apt_add,
        m_subp,
        m_setup_apt_proxy,
        m_update_sources_list,
        entitlement,
        capsys,
        caplog_text,
        event,
        packages,
        series,
        file_extension,
        with_pre_install_msg,
    ):
        """On enable add authenticated apt repo and refresh package lists."""
        m_release_info.return_value = mock.MagicMock(series=series)

        pre_install_msgs = ["Some pre-install information", "Some more info"]
        if with_pre_install_msg:
            messaging_patch = mock.patch(
                M_PATH + "RepoEntitlement.messaging",
                new_callable=mock.PropertyMock,
                return_value={"pre_install": pre_install_msgs},
            )
        else:
            messaging_patch = mock.MagicMock()

        expected_apt_calls = []
        expected_emit_calls = [
            mock.call("message_operation", None),
            mock.call("message_operation", None),
        ]
        expected_progress_calls = [
            mock.call("Configuring APT access to Repo Test Class"),
            mock.call("Updating Repo Test Class package lists"),
        ]

        update_sources_list_call_count = 1
        if packages is not None:
            if len(packages) > 0:
                update_sources_list_call_count += 1
                expected_apt_calls.append(
                    mock.call(
                        [
                            "apt-get",
                            "install",
                            "--assume-yes",
                            " ".join(packages),
                        ],
                        capture=True,
                        retry_sleeps=apt.APT_RETRIES,
                        override_env_vars=None,
                    )
                )
                expected_emit_calls = [
                    mock.call("message_operation", None),
                    mock.call("message_operation", None),
                    mock.call(
                        "message_operation",
                        pre_install_msgs if with_pre_install_msg else None,
                    ),
                    mock.call(
                        "info", "Updating standard Ubuntu package lists"
                    ),
                ]
                expected_progress_calls = [
                    mock.call("Configuring APT access to Repo Test Class"),
                    mock.call("Updating Repo Test Class package lists"),
                    mock.call("Installing Repo Test Class packages"),
                ]
        else:
            packages = entitlement.packages

        progress_mock = mock.MagicMock()
        # We patch the type of entitlement because packages is a property
        with mock.patch.object(type(entitlement), "packages", packages):
            with messaging_patch:
                entitlement.enable(progress_mock)

        expected_calls = [
            mock.call(apt.APT_METHOD_HTTPS_FILE),
            mock.call(apt.CA_CERTIFICATES_FILE),
        ]
        assert expected_calls in m_exists.call_args_list
        assert expected_apt_calls == m_subp.call_args_list
        assert (
            update_sources_list_call_count == m_update_sources_list.call_count
        )
        add_apt_calls = [
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-repotest.{}".format(
                    file_extension
                ),
                "http://REPOTEST/ubuntu",
                "repotest-token",
                ["xenial"],
                entitlement.repo_key_file,
                entitlement.snapshot_urls,
                False,
            )
        ]
        assert add_apt_calls == m_apt_add.call_args_list
        assert 1 == m_setup_apt_proxy.call_count
        assert expected_emit_calls == progress_mock.emit.call_args_list
        assert expected_progress_calls == progress_mock.progress.call_args_list

    @mock.patch("uaclient.apt.setup_apt_proxy")
    @mock.patch(M_PATH + "system.subp")
    def test_failed_install_removes_apt_config_and_packages(
        self, m_subp, _m_setup_apt_proxy, entitlement
    ):
        def fake_subp(args, *other_args, **kwargs):
            if "install" in args:
                raise exceptions.ProcessExecutionError(args)

        m_subp.side_effect = fake_subp

        packages = ["fake_pkg", "and_another"]
        with mock.patch.object(entitlement, "setup_apt_config"):
            with mock.patch.object(
                entitlement, "can_enable", return_value=(True, None)
            ):
                with mock.patch.object(
                    type(entitlement), "packages", packages
                ):
                    with pytest.raises(exceptions.UbuntuProError) as excinfo:
                        with mock.patch.object(
                            entitlement, "remove_apt_config"
                        ) as m_rac:
                            entitlement.enable(mock.MagicMock())

        assert (
            "Unexpected APT error.\n"
            "Invalid command specified '['apt-get', 'install', "
            "'--assume-yes', 'fake_pkg', 'and_another']'.\n"
            "See /var/log/ubuntu-advantage.log"
        ) == excinfo.value.msg
        assert 1 == m_rac.call_count


class TestPerformEnable:
    @pytest.mark.parametrize(
        [
            "supports_access_only",
            "access_only",
            "expected_setup_apt_calls",
            "expected_install_calls",
        ],
        [
            (
                False,
                False,
                [mock.call(mock.ANY)],
                [mock.call(mock.ANY)],
            ),
            (
                False,
                True,
                [mock.call(mock.ANY)],
                [mock.call(mock.ANY)],
            ),
            (
                True,
                False,
                [mock.call(mock.ANY)],
                [mock.call(mock.ANY)],
            ),
            (
                True,
                True,
                [mock.call(mock.ANY)],
                [],
            ),
        ],
    )
    @mock.patch(M_PATH + "RepoEntitlement._check_for_reboot_msg")
    @mock.patch(M_PATH + "RepoEntitlement.install_packages")
    @mock.patch(M_PATH + "RepoEntitlement.setup_apt_config")
    def test_perform_enable(
        self,
        m_setup_apt_config,
        m_install_packages,
        m_check_for_reboot_msg,
        supports_access_only,
        access_only,
        expected_setup_apt_calls,
        expected_install_calls,
        entitlement_factory,
    ):
        with mock.patch.object(
            RepoTestEntitlement, "supports_access_only", supports_access_only
        ):
            entitlement = entitlement_factory(
                RepoTestEntitlement,
                affordances={"series": ["xenial"]},
                access_only=access_only,
            )
            assert entitlement._perform_enable(mock.MagicMock()) is True
            assert (
                m_setup_apt_config.call_args_list == expected_setup_apt_calls
            )
            assert m_install_packages.call_args_list == expected_install_calls


class TestRepoPerformDisable:
    @pytest.mark.parametrize("purge_value", (True, False))
    @mock.patch(M_PATH + "apt.get_installed_packages_by_origin")
    @mock.patch(M_PATH + "apt.get_remote_versions_for_package")
    @mock.patch(M_PATH + "RepoEntitlement.prompt_for_purge")
    @mock.patch(M_PATH + "RepoEntitlement.execute_removal")
    @mock.patch(M_PATH + "RepoEntitlement.execute_reinstall")
    @mock.patch(M_PATH + "RepoEntitlement.remove_apt_config")
    def test_purge_functions_are_called(
        self,
        m_remove_apt_config,
        m_execute_reinstall,
        m_execute_removal,
        m_prompt_for_purge,
        m_get_remote_versions,
        m_get_installed_packages,
        purge_value,
        entitlement_factory,
    ):

        m_packages = []
        for i in range(1, 6):
            package = mock.MagicMock()
            type(package).name = mock.PropertyMock(return_value=str(i))
            m_packages.append(package)

        m_get_installed_packages.return_value = m_packages

        def return_alternatives(p, exclude_origin):
            if not int(p.name) % 2:
                return [int(p.name)]
            return []

        m_get_remote_versions.side_effect = return_alternatives

        with mock.patch.object(RepoTestEntitlement, "origin", "TestOrigin"):
            entitlement = entitlement_factory(
                RepoTestEntitlement,
                affordances={"series": ["xenial"]},
                purge=purge_value,
            )
            assert entitlement._perform_disable(mock.MagicMock()) is True
            assert m_remove_apt_config.call_args_list == [mock.call(mock.ANY)]

            if purge_value:
                assert m_get_installed_packages.call_args_list == [
                    mock.call("TestOrigin")
                ]
                assert m_get_remote_versions.call_args_list == [
                    mock.call(m_packages[0], exclude_origin="TestOrigin"),
                    mock.call(m_packages[1], exclude_origin="TestOrigin"),
                    mock.call(m_packages[2], exclude_origin="TestOrigin"),
                    mock.call(m_packages[3], exclude_origin="TestOrigin"),
                    mock.call(m_packages[4], exclude_origin="TestOrigin"),
                ]
                assert m_prompt_for_purge.call_args_list == [
                    mock.call(
                        [m_packages[0], m_packages[2], m_packages[4]],
                        [(m_packages[1], 2), (m_packages[3], 4)],
                        mock.ANY,
                    )
                ]
                assert m_execute_removal.call_args_list == [
                    mock.call([m_packages[0], m_packages[2], m_packages[4]])
                ]
                assert m_execute_reinstall.call_args_list == [
                    mock.call([(m_packages[1], 2), (m_packages[3], 4)])
                ]
            else:
                assert m_get_installed_packages.call_args_list == []
                assert m_get_remote_versions.call_args_list == []
                assert m_prompt_for_purge.call_args_list == []
                assert m_execute_removal.call_args_list == []
                assert m_execute_reinstall.call_args_list == []

    @pytest.mark.parametrize(
        "repo_cls",
        (
            (RepoTestEntitlement),
            (RepoTestEntitlementRepoWithRemovePackages),
        ),
    )
    @mock.patch("uaclient.system.should_reboot", return_value=False)
    def test_disable_on_can_disable_true_removes_apt_config_and_packages(
        self,
        _m_should_reboot,
        repo_cls,
        entitlement_factory,
    ):
        """When can_disable, disable removes apt config and packages."""
        entitlement = entitlement_factory(repo_cls)

        if isinstance(entitlement, RepoTestEntitlementRepoWithRemovePackages):
            mock_remove_packages = mock.patch.object(
                entitlement, "remove_packages"
            )
        else:
            mock_remove_packages = mock.MagicMock()

        with mock.patch.object(
            entitlement, "can_disable", return_value=(True, None)
        ):
            with mock.patch.object(
                entitlement, "remove_apt_config"
            ) as m_remove_apt_config:
                with mock_remove_packages as m_remove_packages:
                    assert entitlement.disable(mock.MagicMock())

        assert [mock.call(mock.ANY)] == m_remove_apt_config.call_args_list

        if isinstance(entitlement, RepoTestEntitlementRepoWithRemovePackages):
            assert [mock.call()] == m_remove_packages.call_args_list
        else:
            assert 0 == m_remove_packages.call_count


class TestPurge:
    # This circular import should NOT exist, test_apt needs a refactor.
    from uaclient.tests.test_apt import mock_package, mock_version

    packages_to_remove = [
        mock_package("remove1", mock_version("1.0")),
        mock_package("remove2", mock_version("1.0")),
    ]
    packages_to_reinstall = [
        (mock_package("reinstall1", mock_version("1.0")), mock_version("2.0")),
        (mock_package("reinstall2", mock_version("1.0")), mock_version("2.0")),
    ]
    mock_kernel_package = [mock_package("linux-image-1.2.3-4-fake")]

    def test_purge_kernel_check_true_if_no_kernel(self, entitlement_factory):
        entitlement = entitlement_factory(
            RepoTestEntitlement,
            affordances={"series": ["xenial"]},
            purge=True,
        )

        assert (
            entitlement.purge_kernel_check(
                self.packages_to_remove, mock.MagicMock()
            )
            is True
        )

    @mock.patch(
        M_PATH + "system.get_installed_ubuntu_kernels", return_value=[]
    )
    @mock.patch(M_PATH + "system.get_kernel_info")
    def test_purge_kernel_check_false_when_no_other_kernels(
        self, _m_kernel_info, _m_installed_kernels, entitlement_factory
    ):
        entitlement = entitlement_factory(
            RepoTestEntitlement,
            affordances={"series": ["xenial"]},
            purge=True,
        )

        mock_progress = mock.MagicMock()
        assert (
            entitlement.purge_kernel_check(
                self.mock_kernel_package, mock_progress
            )
            is False
        )
        assert (
            mock.call(
                "info",
                (
                    "No other valid Ubuntu kernel was found in the system.\n"
                    "Removing the package would potentially make the system "
                    "unbootable.\n"
                    "Aborting.\n"
                ),
            )
            in mock_progress.emit.call_args_list
        )

    @pytest.mark.parametrize("prompt_answer", (True, False))
    @pytest.mark.parametrize(
        "current_kernel", ("1.2.3-4-fake", "2.3.4-5-fake")
    )
    @mock.patch(
        M_PATH + "system.get_installed_ubuntu_kernels",
        return_value=["2.3.4-5-fake"],
    )
    @mock.patch(M_PATH + "system.get_kernel_info")
    def test_purge_kernel_check_prompts_if_other_kernels(
        self,
        m_kernel_info,
        _m_installed_kernels,
        prompt_answer,
        current_kernel,
        entitlement_factory,
    ):
        m_kernel_info.return_value.uname_release = current_kernel
        entitlement = entitlement_factory(
            RepoTestEntitlement,
            affordances={"series": ["xenial"]},
            purge=True,
        )

        mock_progress = mock.MagicMock()

        def fake_emit(name, payload):
            if name == "message_operation" and prompt_answer is False:
                raise exceptions.PromptDeniedError()

        mock_progress.emit.side_effect = fake_emit

        expect_raises = (
            does_not_raise()
            if prompt_answer
            else pytest.raises(exceptions.PromptDeniedError)
        )
        with expect_raises:
            entitlement.purge_kernel_check(
                self.mock_kernel_package, mock_progress
            )

        assert (
            mock.call(
                "info",
                messages.PURGE_KERNEL_REMOVAL.format(
                    service="Repo Test Class"
                ),
            )
            in mock_progress.emit.call_args_list
        )
        assert (
            mock.call(
                "info",
                messages.PURGE_CURRENT_KERNEL.format(
                    kernel_version=current_kernel
                ),
            )
            in mock_progress.emit.call_args_list
        )

    @pytest.mark.parametrize(
        ["remove", "reinstall", "expected_print", "expected_progress_emits"],
        (
            (
                packages_to_remove,
                packages_to_reinstall,
                [
                    mock.call(["remove1", "remove2"]),
                    mock.call(["reinstall1", "reinstall2"]),
                ],
                [
                    mock.call("info", messages.WARN_PACKAGES_REMOVAL),
                    mock.call("info", "  remove1 remove2\n"),
                    mock.call("info", messages.WARN_PACKAGES_REINSTALL),
                    mock.call("info", "  reinstall1 reinstall2\n"),
                    mock.call(
                        "message_operation",
                        [(mock.ANY, {"msg": messages.PROCEED_YES_NO})],
                    ),
                ],
            ),
            (
                packages_to_remove,
                [],
                [mock.call(["remove1", "remove2"])],
                [
                    mock.call("info", messages.WARN_PACKAGES_REMOVAL),
                    mock.call("info", "  remove1 remove2\n"),
                    mock.call(
                        "message_operation",
                        [(mock.ANY, {"msg": messages.PROCEED_YES_NO})],
                    ),
                ],
            ),
            (
                [],
                packages_to_reinstall,
                [mock.call(["reinstall1", "reinstall2"])],
                [
                    mock.call("info", messages.WARN_PACKAGES_REINSTALL),
                    mock.call("info", "  reinstall1 reinstall2\n"),
                    mock.call(
                        "message_operation",
                        [(mock.ANY, {"msg": messages.PROCEED_YES_NO})],
                    ),
                ],
            ),
            ([], [], [], []),
        ),
    )
    def test_prompt_for_purge(
        self,
        remove,
        reinstall,
        expected_print,
        expected_progress_emits,
        entitlement_factory,
    ):
        entitlement = entitlement_factory(
            RepoTestEntitlement,
            affordances={"series": ["xenial"]},
            purge=True,
        )

        mock_progress = mock.MagicMock()
        assert (
            entitlement.prompt_for_purge(remove, reinstall, mock_progress)
            is True
        )
        assert mock_progress.emit.call_args_list == expected_progress_emits

    @pytest.mark.parametrize(
        "remove,expected_remove",
        (
            (
                packages_to_remove,
                [
                    mock.call(
                        ["remove1", "remove2"],
                        messages.UNINSTALLING_PACKAGES_FAILED.format(
                            packages=["remove1", "remove2"]
                        ),
                    ),
                ],
            ),
            (
                [],
                [],
            ),
        ),
    )
    @mock.patch(M_PATH + "apt.purge_packages")
    @mock.patch(M_PATH + "apt.get_installed_packages_names")
    def test_execute_removal(
        self,
        m_installed_packages,
        m_apt_purge,
        remove,
        expected_remove,
        entitlement_factory,
    ):
        m_installed_packages.return_value = ["remove1", "remove2"]
        entitlement = entitlement_factory(
            RepoTestEntitlement,
            affordances={"series": ["xenial"]},
            purge=True,
        )
        entitlement.execute_removal(remove)

        assert m_apt_purge.call_args_list == expected_remove

    @pytest.mark.parametrize(
        "reinstall,expected_install",
        (
            (
                packages_to_reinstall,
                [
                    mock.call(
                        ["reinstall1=2.0", "reinstall2=2.0"],
                        apt_options=[
                            "--allow-downgrades",
                            '-o Dpkg::Options::="--force-confdef"',
                            '-o Dpkg::Options::="--force-confold"',
                        ],
                        override_env_vars={
                            "DEBIAN_FRONTEND": "noninteractive"
                        },
                    ),
                ],
            ),
            ([], []),
        ),
    )
    @mock.patch(M_PATH + "apt.run_apt_install_command")
    @mock.patch(M_PATH + "apt.get_installed_packages_names")
    def test_execute_reinstall(
        self,
        m_installed_packages,
        m_apt_install,
        reinstall,
        expected_install,
        entitlement_factory,
    ):
        m_installed_packages.return_value = ["reinstall1", "reinstall2"]
        entitlement = entitlement_factory(
            RepoTestEntitlement,
            affordances={"series": ["xenial"]},
            purge=True,
        )
        entitlement.execute_reinstall(reinstall)

        assert m_apt_install.call_args_list == expected_install


class TestUpdateAptAuth:
    @mock.patch(M_PATH + "apt.add_apt_auth_conf_entry")
    def test_update_apt_auth_self_snapshot_urls(
        self, m_add_apt_auth_conf_entry, entitlement_factory
    ):
        entitlement = entitlement_factory(
            RepoTestEntitlement,
            directives={
                "aptURL": "http://REPOTEST",
                "aptSnapshotURLs": ["one", "two", "three"],
            },
        )
        entitlement.update_apt_auth()
        assert m_add_apt_auth_conf_entry.call_args_list == [
            mock.call("http://REPOTEST/ubuntu", "bearer", "repotest-token"),
            mock.call("one", "bearer", "repotest-token"),
            mock.call("two", "bearer", "repotest-token"),
            mock.call("three", "bearer", "repotest-token"),
        ]

    @mock.patch(M_PATH + "apt.add_apt_auth_conf_entry")
    def test_update_apt_auth_override_snapshot_urls(
        self, m_add_apt_auth_conf_entry, entitlement_factory
    ):
        entitlement = entitlement_factory(RepoTestEntitlement)
        entitlement.update_apt_auth(
            override_snapshot_urls=["one", "two", "three"]
        )
        assert m_add_apt_auth_conf_entry.call_args_list == [
            mock.call("http://REPOTEST/ubuntu", "bearer", "repotest-token"),
            mock.call("one", "bearer", "repotest-token"),
            mock.call("two", "bearer", "repotest-token"),
            mock.call("three", "bearer", "repotest-token"),
        ]


class TestRemoveAptConfig:
    def test_missing_aptURL(self, entitlement_factory):
        # Make aptURL missing
        entitlement = entitlement_factory(RepoTestEntitlement, directives={})

        with pytest.raises(exceptions.MissingAptURLDirective) as excinfo:
            entitlement.remove_apt_config(mock.MagicMock())

        assert "repotest" in str(excinfo.value)

    @mock.patch(M_PATH + "apt.remove_auth_apt_repo")
    @mock.patch(M_PATH + "apt.remove_apt_list_files")
    @mock.patch(M_PATH + "apt.run_apt_update_command")
    def test_apt_get_update_called(
        self, m_run_apt_update_command, _m_apt1, _m_apt2, entitlement
    ):
        entitlement.remove_apt_config(mock.MagicMock())

        assert mock.call() in m_run_apt_update_command.call_args_list

    @mock.patch(M_PATH + "apt.remove_auth_apt_repo")
    @mock.patch(M_PATH + "apt.remove_apt_list_files")
    @mock.patch(M_PATH + "apt.run_apt_command")
    @mock.patch(M_PATH + "system.get_release_info")
    def test_disable_removes_all_apt_config(
        self,
        m_get_release_info,
        _m_run_apt_command,
        m_remove_apt_list_files,
        m_remove_auth_apt_repo,
        entitlement_factory,
    ):
        """Remove all APT config when disable_apt_auth_only is False"""
        m_get_release_info.return_value = mock.MagicMock(series="xenial")

        entitlement = entitlement_factory(
            RepoTestEntitlement,
            affordances={"series": ["xenial"]},
        )
        entitlement.remove_apt_config(mock.MagicMock())

        assert [
            mock.call("http://REPOTEST/ubuntu", "xenial")
        ] == m_remove_apt_list_files.call_args_list
        assert [
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-repotest.list",
                "http://REPOTEST/ubuntu",
                [],
                "test.gpg",
            )
        ] == m_remove_auth_apt_repo.call_args_list

    @mock.patch(M_PATH + "system.ensure_file_absent")
    @mock.patch(M_PATH + "apt.remove_auth_apt_repo")
    @mock.patch(M_PATH + "apt.remove_apt_list_files")
    @mock.patch(M_PATH + "apt.run_apt_command")
    @mock.patch(M_PATH + "system.get_release_info")
    @mock.patch(M_PATH + "contract.apply_contract_overrides")
    def test_repo_pin_priority_int_removes_apt_preferences(
        self,
        _m_contract_overrides,
        m_get_release_info,
        _m_run_apt_command,
        _m_remove_apt_list_files,
        _m_remove_auth_apt_repo,
        m_ensure_file_absent,
        entitlement_factory,
    ):
        """Remove apt preferences file when repo_pin_priority is an int."""
        m_get_release_info.return_value = mock.MagicMock(series="xenial")

        entitlement = entitlement_factory(
            RepoTestEntitlementRepoWithPin, affordances={"series": ["xenial"]}
        )

        assert 1000 == entitlement.repo_pin_priority
        entitlement.remove_apt_config(mock.MagicMock())
        assert [
            mock.call("/etc/apt/preferences.d/ubuntu-repotest")
        ] == m_ensure_file_absent.call_args_list


class TestSetupAptConfig:
    @pytest.mark.parametrize(
        "repo_cls,directives,exc_cls,err_msg",
        (
            (
                RepoTestEntitlement,
                {},
                exceptions.UbuntuProError,
                "Ubuntu Pro server provided no aptKey directive for"
                " repotest",
            ),
            (
                RepoTestEntitlement,
                {"aptKey": "somekey"},
                exceptions.MissingAptURLDirective,
                "repotest",
            ),
            (
                RepoTestEntitlement,
                {"aptKey": "somekey", "aptURL": "someURL"},
                exceptions.UbuntuProError,
                "Ubuntu Pro server provided no suites directive for"
                " repotest",
            ),
        ),
    )
    @mock.patch("uaclient.apt.setup_apt_proxy")
    def test_missing_directives(
        self,
        _setup_apt_proxy,
        repo_cls,
        directives,
        exc_cls,
        err_msg,
        entitlement_factory,
    ):
        """Raise an error when missing any required directives."""
        entitlement = entitlement_factory(repo_cls, directives=directives)

        with pytest.raises(exc_cls) as excinfo:
            entitlement.setup_apt_config(mock.MagicMock())
        assert err_msg in str(excinfo.value)

    @pytest.mark.parametrize("enable_by_default", (True, False))
    @mock.patch("uaclient.apt.setup_apt_proxy")
    @mock.patch(M_CONTRACT_PATH + "get_resource_machine_access")
    @mock.patch(M_PATH + "apt.add_auth_apt_repo")
    @mock.patch(M_PATH + "apt.run_apt_command")
    def test_setup_apt_config_request_machine_access_when_no_resource_token(
        self,
        run_apt_command,
        add_auth_apt_repo,
        _m_get_resource_machine_access,
        _setup_apt_proxy,
        enable_by_default,
        entitlement_factory,
        caplog_text,
        fake_machine_token_file,
    ):
        """request_machine_access routes when contract lacks resourceToken."""
        entitlement = entitlement_factory(
            RepoTestEntitlement,
            affordances={"series": ["xenial"]},
            obligations={"enableByDefault": enable_by_default},
        )
        machine_token = fake_machine_token_file.machine_token
        # Drop resourceTokens values from base machine-token.
        machine_token["resourceTokens"] = []
        entitlement.setup_apt_config(mock.MagicMock())
        expected_msg = (
            "No resourceToken present in contract for service Repo Test"
            " Class. Using machine token as credentials"
        )
        if enable_by_default:
            assert expected_msg in caplog_text()
            assert 0 == _m_get_resource_machine_access.call_count
        else:
            assert expected_msg not in caplog_text()
            assert [
                mock.call("blah", "repotest")
            ] == _m_get_resource_machine_access.call_args_list

    @mock.patch("uaclient.apt.setup_apt_proxy")
    @mock.patch("os.path.exists", return_value=False)
    @mock.patch("uaclient.apt.add_auth_apt_repo")
    @mock.patch("uaclient.apt.run_apt_command")
    def test_calls_setup_apt_proxy(
        self,
        _m_run_apt_command,
        _m_add_auth_repo,
        _m_exists,
        m_setup_apt_proxy,
        entitlement,
    ):
        """Calls apt.setup_apt_proxy()"""
        entitlement.setup_apt_config(mock.MagicMock())
        assert [
            mock.call(http_proxy=None, https_proxy=None, proxy_scope=None)
        ] == m_setup_apt_proxy.call_args_list

    @mock.patch("uaclient.apt.setup_apt_proxy")
    @mock.patch(M_PATH + "apt.add_auth_apt_repo")
    @mock.patch(M_PATH + "apt.run_apt_install_command")
    @mock.patch(M_PATH + "contract.apply_contract_overrides")
    def test_install_prerequisite_packages(
        self,
        _m_contract_overrides,
        m_run_apt_install_command,
        m_add_auth_repo,
        _m_setup_apt_proxy,
        entitlement,
    ):
        """Install apt-transport-https and ca-certificates debs if absent.

        Presence is determined based on checking known files from those debs.
        It avoids a costly round-trip shelling out to call dpkg -l.
        """
        with mock.patch(M_PATH + "exists") as m_exists:
            m_exists.return_value = False
            entitlement.setup_apt_config(mock.MagicMock())
        assert [
            mock.call("/usr/lib/apt/methods/https"),
            mock.call("/usr/sbin/update-ca-certificates"),
        ] == m_exists.call_args_list
        install_call = mock.call(
            packages=["apt-transport-https", "ca-certificates"]
        )
        assert install_call in m_run_apt_install_command.call_args_list

    @mock.patch("uaclient.apt.setup_apt_proxy")
    @mock.patch(M_PATH + "system.get_release_info")
    def test_setup_error_with_repo_pin_priority_and_missing_origin(
        self, m_get_release_info, _setup_apt_proxy, entitlement_factory
    ):
        """Raise error when repo_pin_priority is set and origin is None."""
        m_get_release_info.return_value = mock.MagicMock(series="xenial")
        entitlement = entitlement_factory(
            RepoTestEntitlementRepoWithPin, affordances={"series": ["xenial"]}
        )
        with pytest.raises(exceptions.UbuntuProError) as excinfo:
            entitlement.setup_apt_config(mock.MagicMock())
        assert (
            "Cannot setup apt pin. Empty apt repo origin value for repotest"
            in str(excinfo.value)
        )

    @mock.patch("uaclient.apt.update_sources_list")
    @mock.patch("uaclient.apt.setup_apt_proxy")
    @mock.patch(M_PATH + "apt.add_auth_apt_repo")
    @mock.patch(M_PATH + "apt.run_apt_update_command")
    @mock.patch(M_PATH + "apt.add_ppa_pinning")
    @mock.patch(M_PATH + "system.get_release_info")
    @mock.patch(M_PATH + "contract.apply_contract_overrides")
    def test_setup_with_repo_pin_priority_int_adds_a_pins_repo_apt_preference(
        self,
        _m_apply_overrides,
        m_get_release_info,
        m_add_ppa_pinning,
        m_run_apt_update_command,
        m_add_auth_repo,
        m_update_sources_list,
        _m_setup_apt_proxy,
        entitlement_factory,
    ):
        """When repo_pin_priority is an int, set pin in apt preferences."""
        m_get_release_info.return_value = mock.MagicMock(series="xenial")
        entitlement = entitlement_factory(
            RepoTestEntitlementRepoWithPin, affordances={"series": ["xenial"]}
        )
        entitlement.origin = "RepoTestOrigin"  # don't error on origin = None
        with mock.patch(M_PATH + "exists") as m_exists:
            m_exists.return_value = True  # Skip prerequisite pkg installs
            entitlement.setup_apt_config(mock.MagicMock())
        assert [
            mock.call("/usr/lib/apt/methods/https"),
            mock.call("/usr/sbin/update-ca-certificates"),
        ] == m_exists.call_args_list
        assert [
            mock.call(
                "/etc/apt/preferences.d/ubuntu-repotest",
                "http://REPOTEST",
                "RepoTestOrigin",
                entitlement.repo_pin_priority,
            )
        ] == m_add_ppa_pinning.call_args_list
        assert [] == m_run_apt_update_command.call_args_list
        assert 1 == m_update_sources_list.call_count


class TestCheckAptURLIsApplied:
    @pytest.mark.parametrize("apt_url", (("test"), (None)))
    @mock.patch("uaclient.system.load_file")
    def test_check_apt_url_for_commented_apt_source_file(
        self, m_load_file, apt_url, entitlement
    ):
        m_load_file.return_value = "#test1\n#test2\n"
        assert not entitlement._check_apt_url_is_applied(apt_url)

    @mock.patch("uaclient.system.load_file")
    def test_check_apt_url_when_delta_apt_url_is_none(
        self, m_load_file, entitlement
    ):
        m_load_file.return_value = "test1\n#test2\n"
        assert entitlement._check_apt_url_is_applied(apt_url=None)

    @pytest.mark.parametrize(
        "apt_url,expected", (("test", True), ("blah", False))
    )
    @mock.patch("uaclient.system.load_file")
    def test_check_apt_url_inspects_apt_source_file(
        self, m_load_file, apt_url, expected, entitlement
    ):
        m_load_file.return_value = "test\n#test2\n"
        assert expected == entitlement._check_apt_url_is_applied(apt_url)


class TestApplicationStatus:
    # TODO: Write tests for all functionality
    # 🤔

    def test_missing_aptURL(self, entitlement_factory):
        # Make aptURL missing
        entitlement = entitlement_factory(RepoTestEntitlement, directives={})

        application_status, explanation = entitlement.application_status()

        assert ApplicationStatus.DISABLED == application_status
        assert (
            "Repo Test Class does not have an aptURL directive"
            == explanation.msg
        )

    @pytest.mark.parametrize(
        "policy_url,enabled",
        (
            ("https://esm.ubuntu.com/apps/ubuntu", False),
            ("https://esm.ubuntu.com/ubuntu", True),
        ),
    )
    @mock.patch(M_PATH + "apt.get_apt_cache_policy")
    def test_enabled_status_by_apt_policy(
        self, m_run_apt_policy, policy_url, enabled, entitlement_factory
    ):
        """Report ENABLED when apt-policy lists specific aptURL."""
        entitlement = entitlement_factory(
            RepoTestEntitlement,
            directives={
                "aptURL": "https://esm.ubuntu.com",
                "suites": ["bionic-updates", "bionic-security"],
            },
        )

        policy_lines = [
            "500 {policy_url} bionic-security/main amd64 Packages".format(
                policy_url=policy_url
            ),
            " release v=18.04,o=UbuntuESMApps,...,n=bionic,l=UbuntuESMApps",
            "  origin esm.ubuntu.com",
        ]
        m_run_apt_policy.return_value = "\n".join(policy_lines)

        application_status, explanation = entitlement.application_status()

        if enabled:
            expected_status = ApplicationStatus.ENABLED
            expected_explanation = "Repo Test Class is active"
        else:
            expected_status = ApplicationStatus.DISABLED
            expected_explanation = "Repo Test Class is not configured"
        assert expected_status == application_status
        assert expected_explanation == explanation.msg



import logging
from typing import List

from uaclient import entitlements, util
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.data_types import DataObject, Field, StringDataValue, data_list

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


class Reason(DataObject):
    fields = [
        Field(
            "code",
            StringDataValue,
            doc="Short string that represents the reason",
        ),
        Field(
            "title",
            StringDataValue,
            doc="Longer string describing the reason - possibly translated",
        ),
    ]

    def __init__(self, *, code: str, title: str):
        self.code = code
        self.title = title


class ServiceWithReason(DataObject):
    fields = [
        Field(
            "name",
            StringDataValue,
            doc="Name of the Pro service this item corresponds to",
        ),
        Field(
            "reason",
            Reason,
            doc="Reason that this service is in the list it is in",
        ),
    ]

    def __init__(self, *, name: str, reason: Reason):
        self.name = name
        self.reason = reason


class ServiceWithDependencies(DataObject):
    fields = [
        Field(
            "name",
            StringDataValue,
            doc="Name of the Pro service this item corresponds to",
        ),
        Field(
            "incompatible_with",
            data_list(ServiceWithReason),
            doc=(
                "List of Pro services this service is incompatible with. That"
                " means they cannot be enabled at the same time."
            ),
        ),
        Field(
            "depends_on",
            data_list(ServiceWithReason),
            doc=(
                "List of Pro services this service depends on. The services"
                " in this list must be enabled for this service to be enabled."
            ),
        ),
    ]

    def __init__(
        self,
        *,
        name: str,
        incompatible_with: List[ServiceWithReason],
        depends_on: List[ServiceWithReason]
    ):
        self.name = name
        self.incompatible_with = incompatible_with
        self.depends_on = depends_on


class DependenciesResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "services",
            data_list(ServiceWithDependencies),
            doc="Each Pro service gets an item in this list",
        ),
    ]

    def __init__(self, *, services: List[ServiceWithDependencies]):
        self.services = services


def dependencies() -> DependenciesResult:
    return _dependencies(UAConfig())


def _dependencies(cfg: UAConfig) -> DependenciesResult:
    """
    This endpoint will return a full list of all service dependencies,
    regardless of the current system state. That means it will always return
    the same thing until new services are added, or until we add/remove
    dependencies between services.
    """
    services = []
    for ent_cls in entitlements.ENTITLEMENT_CLASSES:
        ent = ent_cls(cfg)
        incompatible_with = []
        depends_on = []
        for ent_with_reason in ent.incompatible_services:
            incompatible_with.append(
                ServiceWithReason(
                    name=ent_with_reason.entitlement.name,
                    reason=Reason(
                        code=ent_with_reason.named_msg.name,
                        title=ent_with_reason.named_msg.msg,
                    ),
                )
            )
        for ent_with_reason in ent.required_services:
            depends_on.append(
                ServiceWithReason(
                    name=ent_with_reason.entitlement.name,
                    reason=Reason(
                        code=ent_with_reason.named_msg.name,
                        title=ent_with_reason.named_msg.msg,
                    ),
                )
            )
        services.append(
            ServiceWithDependencies(
                name=ent_cls.name,
                incompatible_with=incompatible_with,
                depends_on=depends_on,
            )
        )
    return DependenciesResult(services=services)


endpoint = APIEndpoint(
    version="v1",
    name="ServiceDependencies",
    fn=_dependencies,
    options_cls=None,
)

_doc = {
    "introduced_in": "32",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.pro.services.dependencies.v1 import dependencies
result = dependencies()
""",
    "result_class": DependenciesResult,
    "example_cli": "pro api u.pro.services.dependencies.v1",
    "example_json": """
{
    "services": [
        {
            "name": "one",
            "depends_on": [
                {
                    "name": "zero",
                    "reason": {
                        "code": "one-and-zero",
                        "title": "Service One requires service Zero."
                    }
                },
                ...
            ],
            "incompatible_with": [
                {
                    "name": "two",
                    "reason": {
                        "code": "one-and-two",
                        "title": "Services One and Two are not compatible."
                    }
                },
                ...
            ]
        },
        ...
    ]
}
""",
}



import json
import re
from datetime import datetime
from typing import List

import click

DEFAULT_LOG_FILE = "/var/log/ubuntu-advantage.log"


class LogEntry:
    def __init__(self, log_data):
        self.timestamp = datetime.strptime(log_data[0], "%Y-%m-%dT%H:%M:%S.%f")
        self.level = log_data[1]
        self.module = log_data[2]
        self.called_function = log_data[3]
        self.line_number = log_data[4]
        self.message = log_data[5]
        self.metadata = log_data[6]
        if "exc_info" in self.metadata:
            self.exc_info = self.metadata["exc_info"]

    def format_traceback(self, traceback: str, color: bool = False) -> str:
        lines = traceback.split("\n")
        formatted_traceback = []

        for line in lines:
            line = line.strip()
            if line.startswith("File "):
                match = re.match(r'(.*")(.*)(".*)', line)
                if match and color:
                    formatted_line = f"{match.group(1)}{click.style(match.group(2), fg='cyan')}{match.group(3)}"  # noqa
                else:
                    formatted_line = line
            elif line.startswith("raise "):
                formatted_line = click.style(line, fg="red") if color else line
            else:
                formatted_line = line

            formatted_traceback.append(f"    {formatted_line}")

        return "\n".join(formatted_traceback)

    def format(self, color: bool = False) -> str:
        level_color = {
            "DEBUG": "cyan",
            "WARNING": "yellow",
            "ERROR": "red",
        }.get(self.level)

        formatted = click.style(
            f"[{self.timestamp}] {self.level}",
            fg=level_color if color else None,
        )
        formatted += f" - {self.module}.{self.called_function} (line {self.line_number}):\n"  # noqa
        formatted += f"    Message: {self.message}\n"

        if self.metadata and self.metadata != {}:
            formatted += "    Metadata:\n"
            for key, value in self.metadata.items():
                if key != "exc_info":
                    formatted += f"        {key}: {value}\n"

        if hasattr(self, "exc_info"):
            formatted += "    Exception Info:\n"
            formatted += self.format_traceback(self.exc_info, color)

        return formatted.strip()


class LogProcessor:
    def __init__(self, log_level, log_file=None, pattern=None):
        self.log_level = log_level
        self.log_file = log_file
        self.pattern = pattern

    def process(self) -> List[LogEntry]:
        if not self.log_file:
            self.log_file = DEFAULT_LOG_FILE
        with open(self.log_file, "r") as file:
            logs = [json.loads(line) for line in file.readlines()]
        return self.filter_logs([LogEntry(log_data) for log_data in logs])

    def filter_logs(self, logs: List[LogEntry]) -> List[LogEntry]:
        filtered = logs
        if self.log_level:
            filtered = [log for log in filtered if log.level == self.log_level]
        if self.pattern:
            regex = re.compile(self.pattern)
            filtered = [
                log
                for log in filtered
                if regex.search(log.message)
                or any(
                    regex.search(str(value)) for value in log.metadata.values()
                )
            ]
        return filtered


@click.command()
@click.argument(
    "log_level",
    type=click.Choice(["ERROR", "DEBUG", "WARNING"]),
    required=False,
)
@click.option(
    "-f",
    "--file",
    type=str,
    help="Log file to process",
)
@click.option(
    "-p",
    "--pattern",
    type=str,
    help="Regular expression pattern to filter logs",
)
def main(log_level, file, pattern):
    processor = LogProcessor(log_level, file, pattern)
    filtered_logs = processor.process()

    for log_entry in filtered_logs:
        print(log_entry.format(color=True))
        print("-" * 80)


if __name__ == "__main__":
    main()



from typing import List

from uaclient import messages
from uaclient.entitlements import repo
from uaclient.types import MessagingOperationsDict


class CISEntitlement(repo.RepoEntitlement):

    help_doc_url = messages.urls.USG_DOCS
    name = "cis"
    description = messages.CIS_DESCRIPTION
    help_text = messages.CIS_HELP_TEXT
    repo_key_file = "ubuntu-pro-cis.gpg"
    apt_noninteractive = True
    supports_access_only = True
    origin = "UbuntuCIS"

    @property
    def messaging(self) -> MessagingOperationsDict:
        if self._called_name == "usg":
            return {"post_enable": [messages.CIS_USG_POST_ENABLE]}
        ret = {
            "post_enable": [messages.CIS_POST_ENABLE]
        }  # type: MessagingOperationsDict
        if "usg" in self.valid_names:
            ret["pre_can_enable"] = [messages.CIS_IS_NOW_USG]
        return ret

    @property
    def packages(self) -> List[str]:
        if self._called_name == "usg":
            return ["usg"]
        return super().packages

    @property
    def title(self) -> str:
        if self._called_name == "cis":
            return messages.CIS_TITLE
        return messages.CIS_USG_TITLE



import json
import logging
from typing import Any, Dict, List, NamedTuple, Optional

from uaclient import (
    api,
    config,
    contract,
    entitlements,
    event_logger,
    exceptions,
    lock,
    messages,
    status,
    util,
)
from uaclient.api.u.pro.services.dependencies.v1 import (
    ServiceWithDependencies,
    _dependencies,
)
from uaclient.api.u.pro.services.enable.v1 import (
    EnableOptions,
    EnableResult,
    _enable,
)
from uaclient.api.u.pro.status.enabled_services.v1 import (
    EnabledService,
    _enabled_services,
)
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.cli import cli_util
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.parser import HelpCategory
from uaclient.files import machine_token

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))

_EnableOneServiceResult = NamedTuple(
    "_EnableOneServiceResult",
    [
        ("success", bool),
        ("needs_reboot", bool),
        ("error", Optional[Dict[str, Any]]),
    ],
)


def _auto_enable_services(
    cfg: config.UAConfig,
    variant: str,
    assume_yes: bool,
    json_output,
):
    interactive_only_print = cli_util.create_interactive_only_print_function(
        json_output
    )
    machine_token_file = machine_token.get_machine_token_file(cfg)
    services_to_be_enabled = contract.get_enabled_by_default_services(
        cfg, machine_token_file.entitlements()
    )

    if not services_to_be_enabled:
        interactive_only_print(messages.NO_SERVICES_TO_AUTO_ENABLE)
        return

    enabled_services = _enabled_services(cfg).enabled_services
    all_dependencies = _dependencies(cfg).services

    for enable_by_default_service in services_to_be_enabled:
        _enable_one_service(
            cfg=cfg,
            ent_name=enable_by_default_service.name,
            variant=variant,
            access_only=False,
            assume_yes=assume_yes,
            json_output=json_output,
            extra_args=None,
            enabled_services=enabled_services,
            all_dependencies=all_dependencies,
        )


def _enable_landscape(
    cfg: config.UAConfig,
    access_only: bool,
    extra_args,
    progress_object: Optional[api.AbstractProgress] = None,
):
    """
    Landscape gets special treatment because it currently not supported by our
    enable API. This function is a temporary workaround until we have a proper
    API for enabling landscape, which will happen after Landscape is fully
    integrated with the contracts backend.
    """
    progress = api.ProgressWrapper(progress_object)
    landscape = entitlements.LandscapeEntitlement(
        cfg,
        called_name="landscape",
        access_only=access_only,
        extra_args=extra_args,
    )
    success = False
    fail_reason = None

    try:
        with lock.RetryLock(
            lock_holder="cli.enable._enable_landscape",
        ):
            success, fail_reason = landscape.enable(progress=progress)
    except Exception as e:
        lock.clear_lock_file_if_present()
        raise e

    if not success:
        if fail_reason is not None and fail_reason.message is not None:
            reason = fail_reason.message
        else:
            reason = messages.GENERIC_UNKNOWN_ISSUE
        raise exceptions.EntitlementNotEnabledError(
            service="landscape", reason=reason
        )
    return EnableResult(
        enabled=["landscape"], disabled=[], reboot_required=False, messages=[]
    )


def prompt_for_dependency_handling(
    cfg: config.UAConfig,
    service: str,
    all_dependencies: List[ServiceWithDependencies],
    enabled_services: List[EnabledService],
    called_name: str,
    variant: str,
    service_title: str,
):
    incompatible_services = []
    required_services = []
    enabled_service_names = [s.name for s in enabled_services]

    dependencies = next(
        (s for s in all_dependencies if s.name == service), None
    )
    if dependencies is not None:
        incompatible_services = [
            s.name
            for s in dependencies.incompatible_with
            if s.name in enabled_service_names
        ]
        required_services = [
            s.name
            for s in dependencies.depends_on
            if s.name not in enabled_service_names
        ]

    for incompatible_service in incompatible_services:
        cfg_block_disable_on_enable = util.is_config_value_true(
            config=cfg.cfg,
            path_to_value="features.block_disable_on_enable",
        )
        incompatible_service_title = entitlements.get_title(
            cfg, incompatible_service
        )
        user_msg = messages.INCOMPATIBLE_SERVICE.format(
            service_being_enabled=service_title,
            incompatible_service=incompatible_service_title,
        )
        if cfg_block_disable_on_enable or not util.prompt_for_confirmation(
            msg=user_msg
        ):
            raise exceptions.IncompatibleServiceStopsEnable(
                service_being_enabled=service_title,
                incompatible_service=incompatible_service_title,
            )

    for required_service in required_services:
        required_service_title = entitlements.get_title(cfg, required_service)
        user_msg = messages.REQUIRED_SERVICE.format(
            service_being_enabled=service_title,
            required_service=required_service_title,
        )
        if not util.prompt_for_confirmation(msg=user_msg):
            raise exceptions.RequiredServiceStopsEnable(
                service_being_enabled=service_title,
                required_service=required_service_title,
            )

    variant_enabled = next(
        (
            s
            for s in enabled_services
            if s.name == service
            and s.variant_enabled
            and s.variant_name != variant
        ),
        None,
    )
    if variant_enabled is not None and variant is not None:
        to_be_enabled_title = entitlements.get_title(cfg, service, variant)
        enabled_variant_title = entitlements.get_title(
            cfg, service, variant_enabled.variant_name
        )
        cfg_block_disable_on_enable = util.is_config_value_true(
            config=cfg.cfg,
            path_to_value="features.block_disable_on_enable",
        )
        user_msg = messages.INCOMPATIBLE_SERVICE.format(
            service_being_enabled=to_be_enabled_title,
            incompatible_service=enabled_variant_title,
        )
        if cfg_block_disable_on_enable or not util.prompt_for_confirmation(
            msg=user_msg
        ):
            raise exceptions.IncompatibleServiceStopsEnable(
                service_being_enabled=to_be_enabled_title,
                incompatible_service=enabled_variant_title,
            )


def _print_json_output(
    json_output: bool,
    json_response: Dict[str, Any],
    processed_services: List[str],
    failed_services: List[str],
    errors: List[Dict[str, Any]],
    warnings: List[Dict[str, Any]],
    success: bool,
):
    if json_output:
        processed_services.sort()
        failed_services.sort()

        json_response["result"] = "success" if success else "failure"
        json_response["processed_services"] = processed_services
        json_response["failed_services"] = failed_services
        json_response["errors"] = errors
        json_response["warnings"] = warnings

        print(
            json.dumps(
                json_response,
                cls=util.DatetimeAwareJSONEncoder,
                sort_keys=True,
            )
        )


def _enable_one_service(
    cfg: config.UAConfig,
    ent_name: str,
    variant: str,
    access_only: bool,
    assume_yes: bool,
    json_output: bool,
    extra_args: Optional[List[str]],
    enabled_services: List[EnabledService],
    all_dependencies: List[ServiceWithDependencies],
) -> _EnableOneServiceResult:
    interactive_only_print = cli_util.create_interactive_only_print_function(
        json_output
    )
    ent = entitlements.entitlement_factory(
        cfg,
        ent_name,
        variant=variant,
        access_only=access_only,
        extra_args=extra_args,
    )
    real_name = ent.name
    ent_title = ent.title

    already_enabled = next(
        (
            s
            for s in enabled_services
            if s.name == real_name
            and (
                not variant
                or (s.variant_enabled and s.variant_name == variant)
            )
        ),
        None,
    )
    if already_enabled is not None:
        msg = messages.ALREADY_ENABLED.format(title=ent_title)
        interactive_only_print(msg.msg)
        interactive_only_print(messages.ENABLE_FAILED.format(title=ent_title))
        return _EnableOneServiceResult(
            success=False,
            needs_reboot=False,
            error={
                "type": "service",
                "service": ent_name,
                "message": msg.msg,
                "message_code": msg.name,
            },
        )

    if not assume_yes:
        # this never happens for json output because we assert earlier that
        # assume_yes must be True for json output
        try:
            prompt_for_dependency_handling(
                cfg,
                real_name,
                all_dependencies,
                enabled_services,
                called_name=ent_name,
                variant=variant,
                service_title=ent_title,
            )
        except exceptions.UbuntuProError as e:
            LOG.exception(e)
            interactive_only_print(e.msg)
            interactive_only_print(
                messages.ENABLE_FAILED.format(title=ent_title)
            )
            return _EnableOneServiceResult(
                success=False,
                needs_reboot=False,
                error=None,
            )

    try:
        if json_output:
            progress = None
        else:
            progress = cli_util.CLIEnableDisableProgress(assume_yes=assume_yes)

        if real_name == "landscape":
            enable_result = _enable_landscape(
                cfg,
                access_only,
                extra_args=extra_args,
                progress_object=progress,
            )
        else:
            enable_result = _enable(
                EnableOptions(
                    service=ent_name,
                    variant=variant,
                    access_only=access_only,
                ),
                cfg,
                progress_object=progress,
            )

        status.status(cfg=cfg)  # Update the status cache

        if access_only:
            interactive_only_print(
                messages.ACCESS_ENABLED_TMPL.format(title=ent_title)
            )
        else:
            interactive_only_print(
                messages.ENABLED_TMPL.format(title=ent_title)
            )

        needs_reboot = enable_result.reboot_required
        if needs_reboot:
            interactive_only_print(
                messages.ENABLE_REBOOT_REQUIRED_TMPL.format(
                    operation="install"
                )
            )

        for message in enable_result.messages:
            interactive_only_print(message)

        return _EnableOneServiceResult(
            success=True,
            needs_reboot=needs_reboot,
            error=None,
        )

    except exceptions.EntitlementNotEnabledError as e:
        reason = e.additional_info["reason"]
        err_code = reason["code"]
        err_msg = reason["title"]
        err_info = reason["additional_info"]
        interactive_only_print(err_msg)
        interactive_only_print(messages.ENABLE_FAILED.format(title=ent_title))
        return _EnableOneServiceResult(
            success=False,
            needs_reboot=False,
            error={
                "type": "service",
                "service": ent_name,
                "message": err_msg,
                "message_code": err_code,
                "additional_info": err_info,
            },
        )
    except exceptions.UbuntuProError as e:
        interactive_only_print(e.msg)
        interactive_only_print(messages.ENABLE_FAILED.format(title=ent_title))
        return _EnableOneServiceResult(
            success=False,
            needs_reboot=False,
            error={
                "type": "service",
                "service": ent_name,
                "message": e.msg,
                "message_code": e.msg_code,
                "additional_info": e.additional_info,
            },
        )


@cli_util.verify_json_format_args
@cli_util.assert_root
@cli_util.assert_attached(cli_util._raise_enable_disable_unattached_error)
def action_enable(args, *, cfg, **kwargs) -> int:
    """Perform the enable action on a named entitlement.

    @return: 0 on success, 1 otherwise
    """
    processed_services = []  # type: List[str]
    failed_services = []  # type: List[str]
    errors = []  # type: List[Dict[str, Any]]
    warnings = []

    json_response = {
        "_schema_version": event_logger.JSON_SCHEMA_VERSION,
        "needs_reboot": False,
    }

    json_output = args.format == "json"
    # HACK NOTICE: interactive_only_print here will be a no-op "null_print"
    # function defined above if args.format == "json". We use this function
    # throughout enable for things that should get printed in the normal
    # interactive output so that they don't get printed for the json output.
    interactive_only_print = cli_util.create_interactive_only_print_function(
        json_output
    )

    variant = getattr(args, "variant", "")
    access_only = args.access_only
    assume_yes = args.assume_yes
    auto = args.auto

    if variant and access_only:
        raise exceptions.InvalidOptionCombination(
            option1="--access-only", option2="--variant"
        )

    if variant and auto:
        raise exceptions.InvalidOptionCombination(
            option1="--variant", option2="--auto"
        )

    if access_only and auto:
        raise exceptions.InvalidOptionCombination(
            option1="--access-only", option2="--auto"
        )

    interactive_only_print(messages.REFRESH_CONTRACT_ENABLE)
    try:
        contract.refresh(cfg, verbose=False)
    except (exceptions.ConnectivityError, exceptions.UbuntuProError):
        # Inability to refresh is not a critical issue during enable
        LOG.warning("Failed to refresh contract", exc_info=True)
        warnings.append(
            {
                "type": "system",
                "message": messages.E_REFRESH_CONTRACT_FAILURE.msg,
                "message_code": messages.E_REFRESH_CONTRACT_FAILURE.name,
            }
        )

    if not _is_attached(cfg).is_attached_and_contract_valid:
        expired_err = exceptions.ContractExpiredError()
        interactive_only_print(expired_err.msg)
        errors.append(
            {
                "type": "system",
                "message": expired_err.msg,
                "message_code": expired_err.msg_code,
            }
        )
        _print_json_output(
            json_output,
            json_response,
            processed_services,
            failed_services,
            errors,
            warnings,
            success=False,
        )
        return 1

    if auto:
        return _auto_enable_services(
            cfg=cfg,
            variant="",
            assume_yes=True,
            json_output=json_output,
        )

    names = getattr(args, "service", [])
    if not names:
        raise exceptions.NoServicesToEnable()

    (
        entitlements_found,
        entitlements_not_found,
    ) = entitlements.get_valid_entitlement_names(names, cfg)
    enabled_services = _enabled_services(cfg).enabled_services
    all_dependencies = _dependencies(cfg).services

    ret = True
    for ent_name in entitlements.order_entitlements_for_enabling(
        cfg, entitlements_found
    ):
        result = _enable_one_service(
            cfg,
            ent_name,
            variant,
            access_only,
            assume_yes,
            json_output,
            kwargs.get("extra_args"),
            enabled_services,
            all_dependencies,
        )
        if result.success:
            processed_services.append(ent_name)
            if result.needs_reboot:
                json_response["needs_reboot"] = True
        else:
            ret = False
            failed_services.append(ent_name)
            if result.error is not None:
                errors.append(result.error)

    if entitlements_not_found:
        ret = False
        failed_services += entitlements_not_found
        err = entitlements.create_enable_entitlements_not_found_error(
            entitlements_not_found, cfg=cfg
        )
        interactive_only_print(err.msg)
        errors.append(
            {
                "type": "system",
                "service": None,
                "message": err.msg,
                "message_code": err.msg_code,
                "additional_info": err.additional_info,
            }
        )

    contract_client = contract.UAContractClient(cfg)
    contract_client.update_activity_token()

    _print_json_output(
        json_output,
        json_response,
        processed_services,
        failed_services,
        errors,
        warnings,
        success=ret,
    )

    return 0 if ret else 1


enable_command = ProCommand(
    "enable",
    help=messages.CLI_ROOT_ENABLE,
    description=messages.CLI_ENABLE_DESC,
    action=action_enable,
    help_category=HelpCategory.QUICKSTART,
    help_position=3,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "service",
                    help=messages.CLI_ENABLE_SERVICE.format(
                        options=", ".join(
                            entitlements.valid_services(cfg=config.UAConfig())
                        )
                    ),
                    action="store",
                    nargs="*",
                ),
                ProArgument(
                    "--access-only",
                    help=messages.CLI_ENABLE_ACCESS_ONLY,
                    action="store_true",
                ),
                ProArgument(
                    "--assume-yes",
                    help=messages.CLI_ASSUME_YES.format(command="enable"),
                    action="store_true",
                ),
                ProArgument(
                    "--auto",
                    help=messages.CLI_ENABLE_AUTO,
                    action="store_true",
                ),
                ProArgument(
                    "--beta",
                    help=messages.CLI_ENABLE_BETA,
                    action="store_true",
                ),
                ProArgument(
                    "--format",
                    help=messages.CLI_FORMAT_DESC.format(default="cli"),
                    action="store",
                    choices=["cli", "json"],
                    default="cli",
                ),
                ProArgument(
                    "--variant",
                    help=messages.CLI_ENABLE_VARIANT,
                    action="store",
                ),
            ]
        )
    ],
)



from uaclient import event_logger, exceptions, messages
from uaclient.api.u.pro.attach.auto.full_auto_attach.v1 import (
    FullAutoAttachOptions,
    _full_auto_attach,
)
from uaclient.cli import cli_util
from uaclient.cli.commands import ProCommand
from uaclient.cli.parser import HelpCategory

event = event_logger.get_event_logger()


@cli_util.assert_root
def action_auto_attach(args, *, cfg, **kwargs) -> int:
    try:
        _full_auto_attach(
            FullAutoAttachOptions(),
            cfg=cfg,
            mode=event_logger.EventLoggerMode.CLI,
        )
    except exceptions.ConnectivityError:
        event.info(messages.E_ATTACH_FAILURE.msg)
        return 1
    else:
        cli_util.post_cli_attach(cfg)
        return 0


auto_attach_command = ProCommand(
    "auto-attach",
    help=messages.CLI_ROOT_AUTO_ATTACH,
    description=messages.CLI_AUTO_ATTACH_DESC,
    action=action_auto_attach,
    help_category=HelpCategory.OTHER,
)



import json

from behave import then
from hamcrest import assert_that, equal_to

from features.steps.shell import when_i_run_command


def _get_snap_data(context, snap):
    cmd = (
        "curl -sS --unix-socket /run/snapd.socket http://localhost/v2/snaps/{}"
    )
    when_i_run_command(
        context,
        cmd.format(snap),
        "with sudo",
    )


@then("I check that snap `{snap}` is installed")
def i_check_that_snap_is_installed(context, snap):
    _get_snap_data(context, snap)
    snap_data_json = json.loads(context.process.stdout.strip())
    assert_that(snap_data_json["status-code"], equal_to(200))


@then("I check that snap `{snap}` is not installed")
def i_check_that_snap_is_not_installed(context, snap):
    _get_snap_data(context, snap)
    snap_data_json = json.loads(context.process.stdout.strip())
    assert_that(snap_data_json["status-code"], equal_to(404))



import contextlib
import logging
import socket

import mock
import pytest

from uaclient import defaults, exceptions, messages
from uaclient.cli import _warn_about_output_redirection, main
from uaclient.exceptions import (
    AlreadyAttachedError,
    LockHeldError,
    UnattachedError,
)

M_PATH_UACONFIG = "uaclient.config.UAConfig."


class TestMain:
    @pytest.mark.parametrize(
        "exception,expected_error_msg,expected_log",
        (
            (
                TypeError("'NoneType' object is not subscriptable"),
                messages.UNEXPECTED_ERROR.format(
                    error_msg="'NoneType' object is not subscriptable",
                    log_path="/var/log/ubuntu-advantage.log",
                ),
                "Unhandled exception, please file a bug",
            ),
        ),
    )
    @mock.patch("uaclient.cli.event.info")
    @mock.patch("uaclient.cli.LOG.exception")
    @mock.patch("uaclient.log.setup_cli_logging")
    @mock.patch("uaclient.cli.get_parser")
    def test_errors_handled_gracefully(
        self,
        m_get_parser,
        _m_setup_logging,
        m_log_exception,
        m_event_info,
        event,
        exception,
        expected_error_msg,
        expected_log,
        FakeConfig,
    ):
        m_args = m_get_parser.return_value.parse_args.return_value
        m_args.action.side_effect = exception

        with pytest.raises(SystemExit) as excinfo:
            with mock.patch("sys.argv", ["/usr/bin/ua", "subcmd"]):
                with mock.patch(
                    "uaclient.config.UAConfig",
                    return_value=FakeConfig(),
                ):
                    main()

        exc = excinfo.value
        assert 1 == exc.code
        assert [
            mock.call(info_msg=expected_error_msg.msg, file_type=mock.ANY)
        ] == m_event_info.call_args_list
        assert [mock.call(expected_log)] == m_log_exception.call_args_list

    @pytest.mark.parametrize(
        "exception,expected_log",
        (
            (
                KeyboardInterrupt,
                "KeyboardInterrupt",
            ),
        ),
    )
    @mock.patch("uaclient.cli.LOG.error")
    @mock.patch("uaclient.log.setup_cli_logging")
    @mock.patch("uaclient.cli.get_parser")
    def test_interrupt_errors_handled_gracefully(
        self,
        m_get_parser,
        _m_setup_logging,
        m_log_error,
        exception,
        expected_log,
        FakeConfig,
    ):
        m_args = m_get_parser.return_value.parse_args.return_value
        m_args.action.side_effect = exception

        with pytest.raises(SystemExit) as excinfo:
            with mock.patch("sys.argv", ["/usr/bin/ua", "subcmd"]):
                with mock.patch(
                    "uaclient.config.UAConfig",
                    return_value=FakeConfig(),
                ):
                    main()

        exc = excinfo.value
        assert 1 == exc.code

        assert [mock.call(expected_log)] == m_log_error.call_args_list

    @pytest.mark.parametrize(
        "exception,expected_exit_code",
        [
            (UnattachedError(), 1),
            (AlreadyAttachedError(account_name=mock.MagicMock()), 2),
            (
                LockHeldError(
                    pid="123",
                    lock_request="pro reboot-cmds",
                    lock_holder="pro auto-attach",
                ),
                1,
            ),
        ],
    )
    @mock.patch("uaclient.cli.event.info")
    @mock.patch("uaclient.cli.LOG.error")
    @mock.patch("uaclient.log.setup_cli_logging")
    @mock.patch("uaclient.cli.get_parser")
    def test_user_facing_error_handled_gracefully(
        self,
        m_get_parser,
        _m_setup_logging,
        m_log_error,
        m_event_info,
        event,
        exception,
        expected_exit_code,
    ):
        m_args = m_get_parser.return_value.parse_args.return_value
        m_args.action.side_effect = exception
        expected_msg = exception.msg

        with pytest.raises(SystemExit) as excinfo:
            main(["some", "args"])

        exc = excinfo.value
        assert expected_exit_code == exc.code

        assert [
            mock.call(info_msg=expected_msg, file_type=mock.ANY)
        ] == m_event_info.call_args_list
        assert [mock.call(expected_msg)] == m_log_error.call_args_list

    @pytest.mark.parametrize(
        ["error_url", "expected_log_call"],
        (
            (
                "http://nowhere.com",
                mock.call(
                    "Failed to access URL: %s",
                    "http://nowhere.com",
                    exc_info=mock.ANY,
                ),
            ),
        ),
    )
    @mock.patch("uaclient.cli.event.info")
    @mock.patch("uaclient.cli.LOG.exception")
    @mock.patch("uaclient.log.setup_cli_logging")
    @mock.patch("uaclient.cli.get_parser")
    def test_url_error_handled_gracefully(
        self,
        m_get_parser,
        _m_setup_logging,
        m_log_exception,
        m_event_info,
        error_url,
        expected_log_call,
    ):
        m_args = m_get_parser.return_value.parse_args.return_value
        m_args.action.side_effect = exceptions.ConnectivityError(
            cause=socket.gaierror(-2, "Name or service not known"),
            url=error_url,
        )

        with pytest.raises(SystemExit) as excinfo:
            main(["some", "args"])

        exc = excinfo.value
        assert 1 == exc.code

        assert [
            mock.call(
                info_msg=messages.E_CONNECTIVITY_ERROR.format(
                    url=error_url,
                    cause_error="[Errno -2] Name or service not known",
                ).msg,
                file_type=mock.ANY,
            )
        ] == m_event_info.call_args_list
        assert [expected_log_call] == m_log_exception.call_args_list

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @mock.patch("uaclient.log.setup_cli_logging")
    @mock.patch("uaclient.cli.get_parser")
    def test_command_line_is_logged(
        self, _m_get_parser, _m_setup_logging, caplog_text
    ):
        main(["some", "args"])

        log = caplog_text()

        assert "['some', 'args']" in log

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @mock.patch("uaclient.log.setup_cli_logging")
    @mock.patch("uaclient.cli.get_parser")
    @mock.patch(
        "uaclient.cli.util.get_pro_environment",
        return_value={"UA_ENV": "YES", "UA_FEATURES_WOW": "XYZ"},
    )
    def test_environment_is_logged(
        self,
        _m_pro_environment,
        _m_get_parser,
        _m_setup_logging,
        caplog_text,
    ):
        main(["some", "args"])

        log = caplog_text()

        assert "UA_ENV=YES" in log
        assert "UA_FEATURES_WOW=XYZ" in log

    @mock.patch("uaclient.log.setup_cli_logging")
    @mock.patch("uaclient.cli.get_parser")
    @mock.patch("uaclient.cli.UAConfig")
    @pytest.mark.parametrize("config_error", [True, False])
    def test_setup_logging_with_defaults(
        self,
        m_config,
        _m_get_parser,
        m_setup_logging,
        config_error,
        tmpdir,
        FakeConfig,
    ):
        log_file = tmpdir.join("file.log")
        cfg = FakeConfig({"log_file": log_file.strpath})
        if not config_error:
            m_config.return_value = cfg
        else:
            m_config.side_effect = OSError("Error reading UAConfig")

        with contextlib.suppress(SystemExit):
            main(["some", "args"])

        expected_setup_logging_calls = [
            mock.call(
                defaults.CONFIG_DEFAULTS["log_level"],
                defaults.CONFIG_DEFAULTS["log_file"],
            ),
        ]

        if not config_error:
            expected_setup_logging_calls.append(
                mock.call(mock.ANY, cfg.log_file),
            )

        assert expected_setup_logging_calls == m_setup_logging.call_args_list

    @pytest.mark.parametrize(
        "command,out_format,is_tty,should_warn",
        (
            ("status", None, True, False),
            ("status", None, False, True),
            ("status", "tabular", True, False),
            ("status", "tabular", False, True),
            ("status", "json", True, False),
            ("status", "json", False, False),
            ("security-status", None, True, False),
            ("security-status", None, False, True),
            ("security-status", "json", True, False),
            ("security-status", "json", False, False),
        ),
    )
    @pytest.mark.parametrize("caplog_text", [logging.WARNING], indirect=True)
    @mock.patch("uaclient.cli.event.info")
    @mock.patch("sys.stdout.isatty")
    def test_status_human_readable_warning(
        self,
        m_tty,
        m_event_info,
        command,
        out_format,
        is_tty,
        should_warn,
        caplog_text,
    ):
        m_tty.return_value = is_tty

        m_args = mock.MagicMock()
        m_args.command = command
        m_args.format = out_format

        _warn_about_output_redirection(m_args)

        if should_warn:
            assert [
                mock.call(
                    messages.WARNING_HUMAN_READABLE_OUTPUT.format(
                        command=command
                    ),
                    file_type=mock.ANY,
                )
            ] == m_event_info.call_args_list
            assert (
                "Not in a tty and human-readable command called"
                in caplog_text()
            )
        else:
            assert [] == m_event_info.call_args_list
            assert (
                "Not in a tty and human-readable command called"
                not in caplog_text()
            )


# There is a fixture for this function to avoid leaking, as it is called in
# the main CLI function. So, instead of importing it directly, we are using
# the reference for the fixture to test it.
class TestWarnAboutNewVersion:
    @pytest.mark.parametrize("new_version", (None, "1.2.3"))
    @mock.patch("uaclient.cli.event.info")
    @mock.patch("uaclient.cli.version.check_for_new_version")
    def test_warn_about_new_version(
        self,
        m_check_version,
        m_event_info,
        new_version,
        _warn_about_new_version,
    ):
        m_check_version.return_value = new_version

        _warn_about_new_version()

        if new_version:
            assert [
                mock.call(
                    messages.WARN_NEW_VERSION_AVAILABLE_CLI.format(
                        version=new_version
                    ),
                    file_type=mock.ANY,
                )
            ] == m_event_info.call_args_list
        else:
            assert [] == m_event_info.call_args_list

    @pytest.mark.parametrize("command", ("api", "status"))
    @pytest.mark.parametrize("out_format", (None, "tabular", "json"))
    @mock.patch("uaclient.cli.event.info")
    @mock.patch(
        "uaclient.cli.version.check_for_new_version", return_value="1.2.3"
    )
    def test_dont_show_for_api_calls(
        self,
        m_check_version,
        m_event_info,
        command,
        out_format,
        _warn_about_new_version,
    ):
        m_check_version.return_value = "1"

        args = mock.MagicMock()
        args.command = command
        args.format = out_format

        if not out_format:
            del args.format

        _warn_about_new_version(args)

        if command != "api" and out_format != "json":
            assert [
                mock.call(
                    messages.WARN_NEW_VERSION_AVAILABLE_CLI.format(
                        version="1"
                    ),
                    file_type=mock.ANY,
                )
            ] == m_event_info.call_args_list
        else:
            assert [] == m_event_info.call_args_list



from typing import List

from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    Field,
    StringDataValue,
    data_list,
)
from uaclient.security_status import get_livepatch_fixed_cves


class LivepatchCVEObject(DataObject):
    fields = [
        Field("name", StringDataValue, doc="Name (ID) of the CVE"),
        Field("patched", BoolDataValue, doc="Livepatch has patched the CVE"),
    ]

    def __init__(self, name: str, patched: bool):
        self.name = name
        self.patched = patched


class LivepatchCVEsResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "fixed_cves",
            data_list(LivepatchCVEObject),
            doc="List of Livepatch patches for the given system",
        ),
    ]

    def __init__(
        self,
        fixed_cves: List[LivepatchCVEObject],
    ):
        self.fixed_cves = fixed_cves


def livepatch_cves() -> LivepatchCVEsResult:
    return _livepatch_cves(UAConfig())


def _livepatch_cves(cfg: UAConfig) -> LivepatchCVEsResult:
    """
    This endpoint lists Livepatch patches for the currently-running kernel.
    """
    return LivepatchCVEsResult(
        fixed_cves=[
            LivepatchCVEObject(name=cve["name"], patched=cve["patched"])
            for cve in get_livepatch_fixed_cves()
        ]
    )


endpoint = APIEndpoint(
    version="v1",
    name="LivepatchCVEs",
    fn=_livepatch_cves,
    options_cls=None,
)

_doc = {
    "introduced_in": "27.12",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.pro.security.status.livepatch_cves.v1 import livepatch_cves

result = livepatch_cves()
""",  # noqa: E501
    "result_class": LivepatchCVEsResult,
    "exceptions": [],
    "example_cli": "pro api u.pro.security.status.livepatch_cves.v1",
    "example_json": """
{
    "fixed_cves":[
        {
            "name": "<CVE Name>",
            "patched": true
        },
        {
            "name": "<Other CVE Name>",
            "patched": false
        },
    ]
}
""",
}



import copy
import logging
import sys
import textwrap
from collections import OrderedDict
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

from uaclient import (
    event_logger,
    exceptions,
    livepatch,
    lock,
    messages,
    util,
    version,
)
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.config import UA_CONFIGURABLE_KEYS, UAConfig
from uaclient.contract import get_available_resources, get_contract_information
from uaclient.defaults import ATTACH_FAIL_DATE_FORMAT, PRINT_WRAP_WIDTH
from uaclient.entitlements import entitlement_factory
from uaclient.entitlements.entitlement_status import (
    ContractStatus,
    UserFacingAvailability,
    UserFacingConfigStatus,
    UserFacingStatus,
)
from uaclient.files import (
    machine_token,
    notices,
    state_files,
    user_config_file,
)
from uaclient.files.notices import Notice
from uaclient.messages import TxtColor

event = event_logger.get_event_logger()
LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


ESSENTIAL = "essential"
STANDARD = "standard"
ADVANCED = "advanced"

# Turns machine-enum value (english) into human value (potentially translated)
# Also colorizes status output for terminal
STATUS_HUMANIZE_COLORIZE = {
    UserFacingStatus.ACTIVE.value: (
        TxtColor.OKGREEN + messages.STATUS_STATUS_ENABLED + TxtColor.ENDC
    ),
    UserFacingStatus.INACTIVE.value: (
        TxtColor.FAIL + messages.STATUS_STATUS_DISABLED + TxtColor.ENDC
    ),
    UserFacingStatus.INAPPLICABLE.value: (
        TxtColor.DISABLEGREY
        + messages.STATUS_STATUS_INAPPLICABLE
        + TxtColor.ENDC
    ),
    UserFacingStatus.UNAVAILABLE.value: (
        TxtColor.DISABLEGREY
        + messages.STATUS_STATUS_UNAVAILABLE
        + TxtColor.ENDC
    ),
    UserFacingStatus.WARNING.value: (
        TxtColor.WARNINGYELLOW + messages.STATUS_STATUS_WARNING + TxtColor.ENDC
    ),
    ContractStatus.ENTITLED.value: (
        TxtColor.OKGREEN + messages.STATUS_ENTITLED_ENTITLED + TxtColor.ENDC
    ),
    ContractStatus.UNENTITLED.value: (
        TxtColor.DISABLEGREY
        + messages.STATUS_ENTITLED_UNENTITLED
        + TxtColor.ENDC
    ),
    ESSENTIAL: TxtColor.OKGREEN
    + messages.STATUS_SUPPORT_ESSENTIAL
    + TxtColor.ENDC,
    STANDARD: TxtColor.OKGREEN
    + messages.STATUS_SUPPORT_STANDARD
    + TxtColor.ENDC,
    ADVANCED: TxtColor.OKGREEN
    + messages.STATUS_SUPPORT_ADVANCED
    + TxtColor.ENDC,
}


STATUS_UNATTACHED_TMPL = "{name: <17}{available: <11}{description}"

STATUS_SIMULATED_TMPL = """\
{name: <17}{available: <11}{entitled: <11}{auto_enabled: <16}{description}"""

STATUS_HEADER = "{name: <17}{entitled: <10}{status: <13}{description}".format(
    name=messages.STATUS_SERVICE,
    entitled=messages.STATUS_ENTITLED,
    status=messages.STATUS_STATUS,
    description=messages.STATUS_DESCRIPTION,
)
# The widths listed below for entitled and status are actually 9 characters
# less than reality because we colorize the values in entitled and status
# columns. Colorizing has an opening and closing set of unprintable characters
# that factor into formats len() calculations
STATUS_TMPL = "{name: <17}{entitled: <19}{status: <22}{description}"
VARIANT_STATUS_TMPL = (
    "{marker} {name: <15}{entitled: <19}{status: <22}{description}"
)

DEFAULT_STATUS = {
    "_doc": "Content provided in json response is currently considered"
    " Experimental and may change",
    "_schema_version": "0.1",
    "version": version.get_version(),
    "machine_id": None,
    "attached": False,
    "effective": None,
    "expires": None,  # TODO Will this break something?
    "origin": None,
    "services": [],
    "execution_status": UserFacingConfigStatus.INACTIVE.value,
    "execution_details": messages.NO_ACTIVE_OPERATIONS,
    "features": {},
    "notices": [],
    "contract": {
        "id": "",
        "name": "",
        "created_at": "",
        "products": [],
        "tech_support_level": UserFacingStatus.INAPPLICABLE.value,
    },
    "account": {
        "name": "",
        "id": "",
        "created_at": "",
        "external_account_ids": [],
    },
    "simulated": False,
}  # type: Dict[str, Any]


def _get_blocked_by_services(ent):
    return [
        {
            "name": (
                service.entitlement.name
                if not service.entitlement.is_variant
                else service.entitlement.variant_name
            ),
            "reason_code": service.named_msg.name,
            "reason": service.named_msg.msg,
        }
        for service in ent.blocking_incompatible_services()
    ]


def _attached_service_status(
    ent, inapplicable_resources, cfg
) -> Dict[str, Any]:
    warning = None
    status_details = ""
    description_override = ent.status_description_override()
    contract_status = ent.contract_status()
    available = "no" if ent.name in inapplicable_resources else "yes"
    variants = {}

    if contract_status == ContractStatus.UNENTITLED:
        ent_status = UserFacingStatus.UNAVAILABLE
    else:
        if ent.name in inapplicable_resources:
            ent_status = UserFacingStatus.INAPPLICABLE
            description_override = inapplicable_resources[ent.name]
        else:
            ent_status, details = ent.user_facing_status()
            if ent_status == UserFacingStatus.WARNING:
                warning = {
                    "code": details.name,
                    "message": details.msg,
                }
            elif details:
                status_details = details.msg

            if ent_status == UserFacingStatus.INAPPLICABLE:
                available = "no"

            if ent.variants:
                variants = {
                    variant_name: _attached_service_status(
                        variant_cls(cfg=cfg),
                        inapplicable_resources,
                        cfg,
                    )
                    for variant_name, variant_cls in ent.variants.items()
                }

    blocked_by = _get_blocked_by_services(ent)

    service_status = {
        "name": ent.presentation_name,
        "description": ent.description,
        "entitled": contract_status.value,
        "status": ent_status.value,
        "status_details": status_details,
        "description_override": description_override,
        "available": available,
        "blocked_by": blocked_by,
        "warning": warning,
    }

    if not ent.is_variant:
        service_status["variants"] = variants

    return service_status


def _attached_status(cfg: UAConfig) -> Dict[str, Any]:
    """Return configuration of attached status as a dictionary."""
    notices.remove(Notice.AUTO_ATTACH_RETRY_FULL_NOTICE)
    notices.remove(Notice.AUTO_ATTACH_RETRY_TOTAL_FAILURE)
    if _is_attached(cfg).is_attached_and_contract_valid:
        notices.remove(Notice.CONTRACT_EXPIRED)

    response = copy.deepcopy(DEFAULT_STATUS)
    machine_token_file = machine_token.get_machine_token_file(cfg)
    machineTokenInfo = machine_token_file.machine_token["machineTokenInfo"]
    contractInfo = machineTokenInfo["contractInfo"]
    tech_support_level = UserFacingStatus.INAPPLICABLE.value
    response.update(
        {
            "machine_id": machineTokenInfo["machineId"],
            "attached": True,
            "origin": contractInfo.get("origin"),
            "notices": notices.list() or [],
            "contract": {
                "id": contractInfo["id"],
                "name": contractInfo["name"],
                "created_at": contractInfo.get("createdAt", ""),
                "products": contractInfo.get("products", []),
                "tech_support_level": tech_support_level,
            },
            "account": {
                "name": machine_token_file.account["name"],
                "id": machine_token_file.account["id"],
                "created_at": machine_token_file.account.get("createdAt", ""),
                "external_account_ids": machine_token_file.account.get(
                    "externalAccountIDs", []
                ),
            },
        }
    )
    if contractInfo.get("effectiveTo"):
        response["expires"] = machine_token_file.contract_expiry_datetime
    if contractInfo.get("effectiveFrom"):
        response["effective"] = contractInfo["effectiveFrom"]

    resources = machine_token_file.machine_token.get("availableResources")
    if not resources:
        resources = get_available_resources(cfg)

    inapplicable_resources = {
        resource["name"]: resource.get("description")
        for resource in sorted(resources, key=lambda x: x.get("name", ""))
        if not resource.get("available")
    }

    for resource in resources:
        try:
            ent = entitlement_factory(cfg=cfg, name=resource.get("name", ""))
        except exceptions.EntitlementNotFoundError:
            continue

        response["services"].append(
            _attached_service_status(ent, inapplicable_resources, cfg)
        )
    response["services"].sort(key=lambda x: x.get("name", ""))

    support = (
        machine_token_file.entitlements().get("support", {}).get("entitlement")
    )
    if support:
        supportLevel = support.get("affordances", {}).get("supportLevel")
        if supportLevel:
            response["contract"]["tech_support_level"] = supportLevel
    return response


def _unattached_status(cfg: UAConfig) -> Dict[str, Any]:
    """Return unattached status as a dict."""

    response = copy.deepcopy(DEFAULT_STATUS)

    resources = get_available_resources(cfg)
    for resource in resources:
        if resource.get("available"):
            available = UserFacingAvailability.AVAILABLE.value
        else:
            available = UserFacingAvailability.UNAVAILABLE.value
        try:
            ent = entitlement_factory(cfg=cfg, name=resource.get("name", ""))

        except exceptions.EntitlementNotFoundError:
            LOG.debug(
                "Ignoring availability of unknown service %s from contract "
                "server",
                resource.get("name", "without a 'name' key"),
            )
            continue

        # FIXME: we need a better generic unattached availability status
        # that takes into account local information.
        if (
            ent.name == "livepatch"
            and livepatch.on_supported_kernel()
            == livepatch.LivepatchSupport.UNSUPPORTED
        ):
            descr_override = ent.status_description_override()
        else:
            descr_override = None

        response["services"].append(
            {
                "name": resource.get("presentedAs", resource["name"]),
                "description": ent.description,
                "description_override": descr_override,
                "available": available,
            }
        )
    response["services"].sort(key=lambda x: x.get("name", ""))

    return response


def _get_config_status(cfg) -> Dict[str, Any]:
    """Return a dict with execution_status, execution_details and notices.

    Values for execution_status will be one of UserFacingConfigStatus
    enum:
        inactive, active, reboot-required
    execution_details will provide more details about that state.
    notices is a list of tuples with label and description items.
    """
    userStatus = UserFacingConfigStatus
    status_val = userStatus.INACTIVE.value
    status_desc = messages.NO_ACTIVE_OPERATIONS
    (lock_pid, lock_holder) = lock.check_lock_info()
    notices_list = notices.list() or []
    if lock_pid > 0:
        status_val = userStatus.ACTIVE.value
        status_desc = messages.LOCK_HELD.format(
            pid=lock_pid, lock_holder=lock_holder
        )
    elif state_files.reboot_cmd_marker_file.is_present:
        status_val = userStatus.REBOOTREQUIRED.value
        operation = "configuration changes"
        status_desc = messages.ENABLE_REBOOT_REQUIRED_TMPL.format(
            operation=operation
        )
    ret = {
        "execution_status": status_val,
        "execution_details": status_desc,
        "notices": notices_list,
        "config_path": cfg.cfg_path,
        "config": cfg.cfg,
        "features": cfg.features,
    }
    # LP: #2004280 maintain backwards compatibility
    ua_config = user_config_file.user_config.public_config.to_dict()
    for key in UA_CONFIGURABLE_KEYS:
        if hasattr(cfg, key) and ua_config[key] is None:
            val = getattr(cfg, key)
            if isinstance(val, Enum):
                val = val.value
            ua_config[key] = val

    ret["config"]["ua_config"] = ua_config

    return ret


def status(cfg: UAConfig, show_all: bool = False) -> Dict[str, Any]:
    """Return status as a dict, using a cache for non-root users

    When unattached, get available resources from the contract service
    to report detailed availability of different resources for this
    machine.

    Write the status-cache when called by root.
    """
    if _is_attached(cfg).is_attached:
        response = _attached_status(cfg)
    else:
        response = _unattached_status(cfg)

    response.update(_get_config_status(cfg))

    if util.we_are_currently_root():
        state_files.status_cache_file.write(response)

    if not show_all:
        available_services = [
            service
            for service in response.get("services", [])
            if service.get("available", "yes") == "yes"
        ]
        response["services"] = available_services

    return response


def _get_entitlement_information(
    entitlements: List[Dict[str, Any]], entitlement_name: str
) -> Dict[str, Any]:
    """Extract information from the entitlements array."""
    for entitlement in entitlements:
        if entitlement.get("type") == entitlement_name:
            return {
                "entitled": "yes" if entitlement.get("entitled") else "no",
                "auto_enabled": (
                    "yes"
                    if entitlement.get("obligations", {}).get(
                        "enableByDefault"
                    )
                    else "no"
                ),
                "affordances": entitlement.get("affordances", {}),
            }
    return {"entitled": "no", "auto_enabled": "no", "affordances": {}}


def simulate_status(
    cfg, token: str, show_all: bool = False
) -> Tuple[Dict[str, Any], int]:
    """Get a status dictionary based on a token.

    Returns a tuple with the status dictionary and an integer value - 0 for
    success, 1 for failure
    """
    ret = 0
    response = copy.deepcopy(DEFAULT_STATUS)

    try:
        contract_information = get_contract_information(cfg, token)
    except exceptions.ContractAPIError as e:
        if hasattr(e, "code") and e.code == 401:
            raise exceptions.AttachInvalidTokenError()
        raise e

    contract_info = contract_information.get("contractInfo", {})
    account_info = contract_information.get("accountInfo", {})

    response.update(
        {
            "contract": {
                "id": contract_info.get("id", ""),
                "name": contract_info.get("name", ""),
                "created_at": contract_info.get("createdAt", ""),
                "products": contract_info.get("products", []),
            },
            "account": {
                "name": account_info.get("name", ""),
                "id": account_info.get("id"),
                "created_at": account_info.get("createdAt", ""),
                "external_account_ids": account_info.get(
                    "externalAccountIDs", []
                ),
            },
            "simulated": True,
        }
    )

    now = datetime.now(timezone.utc)
    if contract_info.get("effectiveTo"):
        response["expires"] = contract_info.get("effectiveTo")
        expiration_datetime = response["expires"]
        delta = expiration_datetime - now
        if delta.total_seconds() <= 0:
            message = messages.E_ATTACH_FORBIDDEN_EXPIRED.format(
                contract_id=response["contract"]["id"],
                date=expiration_datetime.strftime(ATTACH_FAIL_DATE_FORMAT),
            )
            event.error(error_msg=message.msg, error_code=message.name)
            event.info(
                messages.STATUS_TOKEN_NOT_VALID + "\n" + message.msg + "\n"
            )
            ret = 1
    if contract_info.get("effectiveFrom"):
        response["effective"] = contract_info.get("effectiveFrom")
        effective_datetime = response["effective"]
        delta = now - effective_datetime
        if delta.total_seconds() <= 0:
            message = messages.E_ATTACH_FORBIDDEN_NOT_YET.format(
                contract_id=response["contract"]["id"],
                date=effective_datetime.strftime(ATTACH_FAIL_DATE_FORMAT),
            )
            event.error(error_msg=message.msg, error_code=message.name)
            event.info(
                messages.STATUS_TOKEN_NOT_VALID + "\n" + message.msg + "\n"
            )
            ret = 1

    resources = get_available_resources(cfg)
    inapplicable_resources = [
        resource["name"]
        for resource in sorted(resources, key=lambda x: x["name"])
        if not resource["available"]
    ]

    entitlements = contract_info.get("resourceEntitlements", [])
    for resource in resources:
        entitlement_name = resource.get("name", "")
        try:
            ent = entitlement_factory(cfg=cfg, name=entitlement_name)
        except exceptions.EntitlementNotFoundError:
            continue
        entitlement_information = _get_entitlement_information(
            entitlements, entitlement_name
        )
        response["services"].append(
            {
                "name": resource.get("presentedAs", ent.name),
                "description": ent.description,
                "entitled": entitlement_information["entitled"],
                "auto_enabled": entitlement_information["auto_enabled"],
                "available": (
                    "yes" if ent.name not in inapplicable_resources else "no"
                ),
            }
        )
    response["services"].sort(key=lambda x: x.get("name", ""))

    support = _get_entitlement_information(entitlements, "support")
    if support["entitled"]:
        supportLevel = support["affordances"].get("supportLevel")
        if supportLevel:
            response["contract"]["tech_support_level"] = supportLevel

    response.update(_get_config_status(cfg))

    if not show_all:
        available_services = [
            service
            for service in response.get("services", [])
            if service.get("available", "yes") == "yes"
        ]
        response["services"] = available_services

    return response, ret


def for_human_colorized(string: str) -> str:
    """Return colorized string if using a tty, else original string."""
    return (
        STATUS_HUMANIZE_COLORIZE.get(string, string)
        if sys.stdout.isatty()
        else string
    )


def colorize_commands(commands: List[List[str]]) -> str:
    content = ""
    for cmd in commands:
        if content:
            content += " && "
        content += " ".join(cmd)
    # subtract 4 from print width to account for leading and trailing braces
    # and spaces
    wrapped_content = " \\\n".join(
        textwrap.wrap(
            content, width=(PRINT_WRAP_WIDTH - 4), subsequent_indent="  "
        )
    )
    if "\n" in wrapped_content:
        prefix = "{\n  "
        suffix = "\n}"
    else:
        prefix = "{ "
        suffix = " }"
    return "{color}{prefix}{content}{suffix}{end}".format(
        color=TxtColor.DISABLEGREY,
        prefix=prefix,
        content=wrapped_content,
        suffix=suffix,
        end=TxtColor.ENDC,
    )


def get_section_column_content(
    column_data: List[Tuple[str, str]], header: Optional[str] = None
) -> List[str]:
    """Return a list of content lines to print to console for a section

    Content lines will be center-aligned based on max value length of first
    column.
    """
    content = []
    if header:
        content.append(header)
    template_length = max([len(pair[0]) for pair in column_data])
    if template_length > 0:
        template = "{{:>{}}}: {{}}".format(template_length)
        content.extend([template.format(*pair) for pair in column_data])
    else:
        # Then we have an empty "label" column and only descriptions
        content.extend([pair[1] for pair in column_data])
    return content


def format_expires(expires: Optional[datetime]) -> str:
    if expires is None:
        return messages.STATUS_CONTRACT_EXPIRES_UNKNOWN
    try:
        expires = expires.astimezone()
    except Exception:
        pass
    return expires.strftime("%c %Z")


def format_tabular(status: Dict[str, Any], show_all: bool = False) -> str:
    """Format status dict for tabular output."""
    if not status.get("attached"):
        if status.get("simulated"):
            if not status.get("services", None):
                return messages.STATUS_NO_SERVICES_AVAILABLE

            content = [
                STATUS_SIMULATED_TMPL.format(
                    name=messages.STATUS_SERVICE,
                    available=messages.STATUS_AVAILABLE,
                    entitled=messages.STATUS_ENTITLED,
                    auto_enabled=messages.STATUS_AUTO_ENABLED,
                    description=messages.STATUS_DESCRIPTION,
                )
            ]
            for service in status.get("services", []):
                content.append(STATUS_SIMULATED_TMPL.format(**service))

            return "\n".join(content)

        if not status.get("services", None):
            content = [messages.STATUS_NO_SERVICES_AVAILABLE]
        else:
            content = [
                STATUS_UNATTACHED_TMPL.format(
                    name=messages.STATUS_SERVICE,
                    available=messages.STATUS_AVAILABLE,
                    description=messages.STATUS_DESCRIPTION,
                )
            ]
            for service in status.get("services", []):
                descr_override = service.get("description_override")
                description = (
                    descr_override
                    if descr_override
                    else service.get("description", "")
                )
                available = (
                    messages.STANDALONE_YES
                    if service.get("available") == "yes"
                    else messages.STANDALONE_NO
                )
                content.append(
                    STATUS_UNATTACHED_TMPL.format(
                        name=service.get("name", ""),
                        available=available,
                        description=description,
                    )
                )

        notices = status.get("notices")
        if notices:
            content.append(messages.STATUS_NOTICES)
            content.extend(notices)

        if status.get("features"):
            content.append("\n" + messages.STATUS_FEATURES)
            for key, value in sorted(status.get("features", {}).items()):
                content.append("{}: {}".format(key, value))

        if not show_all:
            content.extend(["", messages.STATUS_ALL_HINT])

        content.extend(["", messages.E_UNATTACHED.msg])
        if (
            livepatch.on_supported_kernel()
            == livepatch.LivepatchSupport.UNSUPPORTED
        ):
            content.extend(
                ["", messages.LIVEPATCH_KERNEL_NOT_SUPPORTED_UNATTACHED]
            )
        return "\n".join(content)

    service_warnings = []
    has_variants = False
    if not status.get("services", None):
        content = [messages.STATUS_NO_SERVICES_AVAILABLE]
    else:
        content = [STATUS_HEADER]
        for service_status in status.get("services", []):
            entitled = service_status.get("entitled", "")
            descr_override = service_status.get("description_override")
            description = (
                descr_override
                if descr_override
                else service_status.get("description", "")
            )
            fmt_args = {
                "name": service_status.get("name", ""),
                "entitled": for_human_colorized(entitled),
                "status": for_human_colorized(
                    service_status.get("status", "")
                ),
                "description": description,
            }
            warning = service_status.get("warning", None)
            if warning is not None:
                warning_message = warning.get("message", None)
                if warning_message is not None:
                    service_warnings.append(warning_message)
            variants = service_status.get("variants")
            if variants and not show_all:
                has_variants = True
                fmt_args["name"] = "{}*".format(fmt_args["name"])

            content.append(STATUS_TMPL.format(**fmt_args))
            if variants and show_all:
                for idx, (_, variant) in enumerate(variants.items()):
                    marker = "├" if idx != len(variants) - 1 else "└"
                    content.append(
                        VARIANT_STATUS_TMPL.format(
                            marker=marker,
                            name=variant.get("name"),
                            entitled=for_human_colorized(
                                variant.get("entitled", "")
                            ),
                            status=for_human_colorized(
                                variant.get("status", "")
                            ),
                            description=variant.get("description", ""),
                        )
                    )

    if has_variants:
        content.append("")
        content.append(messages.STATUS_SERVICE_HAS_VARIANTS)

    if status.get("notices") or len(service_warnings) > 0:
        content.append("")
        content.append(messages.STATUS_NOTICES)
        notices = status.get("notices")
        if notices:
            content.extend(notices)
        if len(service_warnings) > 0:
            content.extend(service_warnings)

    if status.get("features"):
        content.append("\n" + messages.STATUS_FEATURES)
        for key, value in sorted(status.get("features", {}).items()):
            content.append("{}: {}".format(key, value))
    content.append("")

    if not show_all:
        if has_variants:
            content.append(messages.STATUS_ALL_HINT_WITH_VARIANTS)
        else:
            content.append(messages.STATUS_ALL_HINT)

    content.append(
        messages.STATUS_FOOTER_ENABLE_SERVICES_WITH.format(
            command="pro enable <service>"
        )
    )
    pairs = []

    account_name = status.get("account", {}).get("name", "unknown")
    if account_name:
        pairs.append((messages.STATUS_FOOTER_ACCOUNT, account_name))

    contract_name = status.get("contract", {}).get("name", "unknown")
    if contract_name:
        pairs.append((messages.STATUS_FOOTER_SUBSCRIPTION, contract_name))

    if status.get("origin", None) != "free":
        pairs.append(
            (
                messages.STATUS_FOOTER_VALID_UNTIL,
                format_expires(status.get("expires")),
            )
        )
        tech_support_level = status.get("contract", {}).get(
            "tech_support_level", "unknown"
        )
        pairs.append(
            (
                messages.STATUS_FOOTER_SUPPORT_LEVEL,
                for_human_colorized(tech_support_level),
            )
        )

    if pairs:
        content.append("")
        content.extend(get_section_column_content(column_data=pairs))

    return "\n".join(content)


def help(cfg, name):
    """Return help information from an uaclient service as a dict

    :param name: Name of the service for which to return help data.

    :raises: UbuntuProError when no help is available.
    """
    resources = get_available_resources(cfg)
    help_resource = None

    # We are using an OrderedDict here to guarantee
    # that if we need to print the result of this
    # dict, the order of insertion will always be respected
    response_dict = OrderedDict()
    response_dict["name"] = name

    for resource in resources:
        if resource["name"] == name or resource.get("presentedAs") == name:
            try:
                help_ent = entitlement_factory(cfg=cfg, name=resource["name"])
            except exceptions.EntitlementNotFoundError:
                continue
            help_resource = resource
            break

    if help_resource is None:
        raise exceptions.NoHelpContent(name=name)

    if _is_attached(cfg).is_attached:
        service_status = _attached_service_status(help_ent, {}, cfg)
        status_msg = service_status["status"]

        response_dict["entitled"] = service_status["entitled"]
        response_dict["status"] = status_msg

    else:
        if help_resource["available"]:
            available = UserFacingAvailability.AVAILABLE.value
        else:
            available = UserFacingAvailability.UNAVAILABLE.value

        response_dict["available"] = available

    response_dict["help"] = help_ent.help_info
    return response_dict



import json
from enum import Enum
from typing import Callable, Dict, Generic, Optional, Type, TypeVar

from uaclient import exceptions
from uaclient.data_types import DataObject
from uaclient.files.files import UAFile
from uaclient.util import DatetimeAwareJSONDecoder
from uaclient.yaml import parser as yaml_parser
from uaclient.yaml import safe_dump, safe_load


class DataObjectFileFormat(Enum):
    JSON = "json"
    YAML = "yaml"


DOFType = TypeVar("DOFType", bound=DataObject)


class DataObjectFile(Generic[DOFType]):
    def __init__(
        self,
        data_object_cls: Type[DOFType],
        ua_file: UAFile,
        file_format: DataObjectFileFormat = DataObjectFileFormat.JSON,
        preprocess_data: Optional[Callable[[Dict], Dict]] = None,
        optional_type_errors_become_null: bool = False,
    ):
        self.data_object_cls = data_object_cls
        self.ua_file = ua_file
        self.file_format = file_format
        self.preprocess_data = preprocess_data
        self.optional_type_errors_become_null = (
            optional_type_errors_become_null
        )

    def read(self) -> Optional[DOFType]:
        raw_data = self.ua_file.read()
        if raw_data is None:
            return None

        parsed_data = None
        if self.file_format == DataObjectFileFormat.JSON:
            try:
                parsed_data = json.loads(
                    raw_data, cls=DatetimeAwareJSONDecoder
                )
            except json.JSONDecodeError:
                raise exceptions.InvalidFileFormatError(
                    file_name=self.ua_file.path, file_format="json"
                )
        elif self.file_format == DataObjectFileFormat.YAML:
            try:
                parsed_data = safe_load(raw_data)
            except yaml_parser.ParserError:
                raise exceptions.InvalidFileFormatError(
                    file_name=self.ua_file.path, file_format="yaml"
                )

        if parsed_data is None:
            return None

        if self.preprocess_data:
            parsed_data = self.preprocess_data(parsed_data)

        return self.data_object_cls.from_dict(
            parsed_data,
            optional_type_errors_become_null=self.optional_type_errors_become_null,  # noqa: E501
        )

    def write(self, content: DOFType):
        if self.file_format == DataObjectFileFormat.JSON:
            str_content = content.to_json()
        elif self.file_format == DataObjectFileFormat.YAML:
            data = content.to_dict()
            str_content = safe_dump(data, default_flow_style=False)

        self.ua_file.write(str_content)

    def delete(self):
        self.ua_file.delete()

    @property
    def path(self):
        return self.ua_file.path



import logging
import sys

from uaclient import util
from uaclient.messages import BROKEN_YAML_MODULE, MISSING_YAML_MODULE

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))

try:
    import yaml
except ImportError as e:
    LOG.exception(e)
    print(MISSING_YAML_MODULE, file=sys.stderr)
    sys.exit(1)


def safe_load(stream):
    try:
        return yaml.safe_load(stream)
    except AttributeError as e:
        LOG.exception(e)
        print(BROKEN_YAML_MODULE.format(path=yaml.__path__), file=sys.stderr)
        sys.exit(1)


def safe_dump(data, stream=None, **kwargs):
    try:
        return yaml.safe_dump(data, stream, **kwargs)
    except AttributeError as e:
        LOG.exception(e)
        print(BROKEN_YAML_MODULE.format(path=yaml.__path__), file=sys.stderr)
        sys.exit(1)


parser = yaml.parser



#!/usr/bin/env python3

"""
Run configuration operations during system boot.

Some uaclient operations cannot be fully completed by running a single
command. For example, when upgrading uaclient from trusty to xenial,
we may have a livepatch change in the contract, allowing livepatch to be
enabled on xenial. However, during the upgrade we cannot install livepatch on
the system because the running kernel version will not be updated until reboot.

Pro client touches a flag file
/var/lib/ubuntu-advantage/marker-reboot-cmds-required to indicate this script
should run at next boot to process any pending/unresovled config operations.
"""

import logging
import sys

from uaclient import (
    api,
    config,
    contract,
    exceptions,
    http,
    lock,
    log,
    update_contract_info,
    upgrade_lts_contract,
)
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.entitlements.fips import FIPSEntitlement
from uaclient.files import machine_token, notices, state_files

LOG = logging.getLogger("ubuntupro.lib.reboot_cmds")


def fix_pro_pkg_holds(cfg: config.UAConfig):
    status_cache = state_files.status_cache_file.read()
    if not status_cache:
        return
    for service in status_cache.get("services", []):
        if service.get("name") == "fips":
            if service.get("status") == "enabled":
                # fips was enabled, fix the holds
                break
            else:
                # fips was not enabled, don't do anything
                return

    LOG.info("Attempting to remove Ubuntu Pro FIPS package holds")
    fips = FIPSEntitlement(cfg)

    try:
        fips.setup_apt_config(
            progress=api.ProgressWrapper()
        )  # Removes package holds
        LOG.info("Successfully removed Ubuntu Pro FIPS package holds")
    except Exception as e:
        LOG.error(e)
        LOG.warning("Could not remove Ubuntu Pro FIPS package holds")

    try:
        fips.install_packages(
            progress=api.ProgressWrapper(), cleanup_on_failure=False
        )
    except exceptions.UbuntuProError:
        LOG.warning("Failed to install packages at boot: %r", fips.packages)
        raise


def refresh_contract(cfg: config.UAConfig):
    try:
        contract.refresh(cfg)
    except exceptions.ConnectivityError:
        LOG.warning("Failed to refresh contract")
        raise


def handle_unattached_state(cfg: config.UAConfig):
    if not _is_attached(cfg).is_attached:
        LOG.info("Skipping reboot_cmds. Machine is unattached")
        if state_files.reboot_cmd_marker_file.is_present:
            state_files.reboot_cmd_marker_file.delete()
        notices.remove(notices.Notice.REBOOT_SCRIPT_FAILED)
        return 0
    return None


def run_reboot_commands(cfg: config.UAConfig):
    fix_pro_pkg_holds(cfg)
    refresh_contract(cfg)
    upgrade_lts_contract.process_contract_delta_after_apt_lock(cfg)
    state_files.reboot_cmd_marker_file.delete()
    notices.remove(notices.Notice.REBOOT_SCRIPT_FAILED)


def handle_only_series_marker_file(cfg: config.UAConfig):
    """Handle the only_series marker file.
    Checks if the marker file is present,
    validates the release series and
    deletes the marker file if the field is not present.""
    """
    if state_files.only_series_check_marker_file.is_present:
        machine_token_file = machine_token.get_machine_token_file()
        only_series = machine_token_file.only_series
        if only_series:
            update_contract_info.validate_release_series(cfg, only_series)
        else:
            state_files.only_series_check_marker_file.delete()


def main(cfg: config.UAConfig) -> int:
    ret = handle_unattached_state(cfg)
    if ret is not None:
        return ret
    try:
        LOG.info("Running commands on reboot")
        with lock.RetryLock(lock_holder="pro-reboot-cmds"):
            handle_only_series_marker_file(cfg)
            if state_files.reboot_cmd_marker_file.is_present:
                run_reboot_commands(cfg)
            else:
                LOG.info("Skipping reboot_cmds. Marker file not present")
                notices.remove(notices.Notice.REBOOT_SCRIPT_FAILED)
                return 0

    except exceptions.LockHeldError as e:
        LOG.warning("Lock not released. %s", str(e.msg))
        notices.add(notices.Notice.REBOOT_SCRIPT_FAILED)
        return 1
    except exceptions.UbuntuProError as e:
        LOG.error(
            "Error while running commands on reboot: %s, %s", e.msg_code, e.msg
        )
        notices.add(notices.Notice.REBOOT_SCRIPT_FAILED)
        return 1
    except Exception as e:
        LOG.error("Failed running commands on reboot. Error: %s", str(e))
        notices.add(notices.Notice.REBOOT_SCRIPT_FAILED)
        return 1

    LOG.info("Successfully ran all commands on reboot.")
    return 0


if __name__ == "__main__":
    log.setup_journald_logging()
    cfg = config.UAConfig()
    http.configure_web_proxy(cfg.http_proxy, cfg.https_proxy)
    sys.exit(main(cfg=cfg))



"""
Timer used to run all jobs that need to be frequently run on the system
"""

import logging
from datetime import datetime, timedelta, timezone
from typing import Callable, Optional

from uaclient import http, log
from uaclient.config import UAConfig
from uaclient.exceptions import (
    InvalidFileEncodingError,
    InvalidFileFormatError,
)
from uaclient.files import machine_token
from uaclient.files.state_files import (
    AllTimerJobsState,
    TimerJobState,
    timer_jobs_state_file,
)
from uaclient.timer.metering import metering_enabled_resources
from uaclient.timer.update_messaging import update_motd_messages

LOG = logging.getLogger("ubuntupro.timer")
UPDATE_MESSAGING_INTERVAL = 21600  # 6 hours
METERING_INTERVAL = 14400  # 4 hours
CHECK_RELEASE_SERIES_INTERVAL = 86400  # 24 hours


class TimedJob:
    def __init__(
        self,
        name: str,
        job_func: Callable[..., bool],
        default_interval_seconds: int,
    ):
        self.name = name
        self._job_func = job_func
        self._default_interval_seconds = default_interval_seconds

    def run(self, cfg: UAConfig) -> bool:
        """Run a job in a failsafe manner, returning True on success.
        Checks if the job is not disabled before running it.

        :param cfg: UAConfig instance

        :return: A bool True when successfully run. False if ignored
            or in error.
        """
        if not self._should_run(cfg):
            return False

        try:
            LOG.info("Running job: %s", self.name)
            if self._job_func(cfg=cfg):
                LOG.info("Executed job: %s", self.name)
        except Exception as e:
            LOG.error("Error executing job %s: %s", self.name, str(e))
            return False

        return True

    def run_interval_seconds(self, cfg: UAConfig) -> int:
        """Return the run_interval for the job based on config or defaults."""
        configured_interval = getattr(cfg, "{}_timer".format(self.name), None)
        if configured_interval is None:
            return self._default_interval_seconds
        if not isinstance(configured_interval, int) or configured_interval < 0:
            LOG.warning(
                "Invalid value for %s interval found in config. "
                "Default value will be used.",
                self.name,
            )
            return self._default_interval_seconds
        return configured_interval

    def _should_run(self, cfg) -> bool:
        """Verify if the job has a valid (non-zero) interval."""
        return self.run_interval_seconds(cfg) != 0


class MeteringTimedJob(TimedJob):
    def __init__(
        self, job_func: Callable[..., bool], default_interval_seconds: int
    ):
        super().__init__(
            name="metering",
            job_func=job_func,
            default_interval_seconds=default_interval_seconds,
        )

    def run_interval_seconds(self, cfg: UAConfig) -> int:
        """
        Define the run interval for the metering job.

        The contract server can control the time we should make the request
        again. Since the user can also configure the timer interval for this
        job, we will select the greater value between those two choices.
        """
        machine_token_file = machine_token.get_machine_token_file(cfg)
        run_interval_seconds = super().run_interval_seconds(cfg)

        if run_interval_seconds == 0:
            # If the user has disabled the metering job, we should
            # ignore the activity_ping_interval directive
            return 0

        return max(
            machine_token_file.activity_ping_interval or 0,
            super().run_interval_seconds(cfg),
        )


metering_job = MeteringTimedJob(metering_enabled_resources, METERING_INTERVAL)
update_message_job = TimedJob(
    "update_messaging",
    update_motd_messages,
    UPDATE_MESSAGING_INTERVAL,
)


def run_job(
    cfg: UAConfig,
    job: TimedJob,
    current_time: datetime,
    job_status: Optional[TimerJobState],
) -> Optional[TimerJobState]:
    if job_status:
        next_run = job_status.next_run
        if next_run and next_run > current_time:
            return job_status
    if job.run(cfg):
        # Persist last_run and next_run UTC-based times on job success.
        last_run = current_time
        next_run = current_time + timedelta(
            seconds=job.run_interval_seconds(cfg)
        )
        job_status = TimerJobState(next_run=next_run, last_run=last_run)

    return job_status


def run_jobs(cfg: UAConfig, current_time: datetime):
    """Run jobs in order when next_run is before current_time.

    Persist jobs-status with calculated next_run values to aid in timer
    state introspection for jobs which have not yet run.
    """
    jobs_status_obj = None
    # If the file format or encoding is wrong we remove it.
    # After the jobs are executed it will be recreated with the proper data.
    try:
        jobs_status_obj = timer_jobs_state_file.read()
    except (InvalidFileEncodingError, InvalidFileFormatError):
        try:
            timer_jobs_state_file.delete()
        except (OSError, PermissionError) as exception:
            LOG.warning(
                "Error trying to delete invalid jobs-status.json file: %s",
                str(exception),
            )
            return

    if jobs_status_obj is None:
        # We do this for the first run of the timer job, where the file
        # doesn't exist
        jobs_status_obj = AllTimerJobsState(
            metering=None,
            update_messaging=None,
        )

    jobs_status_obj.metering = run_job(
        cfg, metering_job, current_time, jobs_status_obj.metering
    )
    jobs_status_obj.update_messaging = run_job(
        cfg, update_message_job, current_time, jobs_status_obj.update_messaging
    )
    timer_jobs_state_file.write(jobs_status_obj)


if __name__ == "__main__":
    log.setup_journald_logging()

    cfg = UAConfig()
    current_time = datetime.now(timezone.utc)

    http.configure_web_proxy(cfg.http_proxy, cfg.https_proxy)

    run_jobs(cfg=cfg, current_time=current_time)



import logging
import sys
import tarfile
import tempfile

from uaclient import messages
from uaclient.actions import collect_logs
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.parser import HelpCategory
from uaclient.util import replace_top_level_logger_name

PRO_COLLECT_LOGS_FILE = "pro_logs.tar.gz"
LOG = logging.getLogger(replace_top_level_logger_name(__name__))


def action_collect_logs(args, *, cfg, **kwargs):
    output_file = args.output or PRO_COLLECT_LOGS_FILE
    with tempfile.TemporaryDirectory() as output_dir:
        collect_logs(cfg, output_dir)
        try:
            with tarfile.open(output_file, "x:gz") as results:
                results.add(output_dir, arcname="logs/")
        except PermissionError as e:
            LOG.error(e)
            return 1
        except FileExistsError as e:
            LOG.error(e)
            print(
                messages.E_FILE_ALREADY_EXISTS.format(
                    filename=output_file,
                ),
                file=sys.stderr,
            )
            return 1

    return 0


collect_logs_command = ProCommand(
    "collect-logs",
    help=messages.CLI_ROOT_COLLECT_LOGS,
    description=messages.CLI_COLLECT_LOGS_DESC,
    preserve_description=True,
    action=action_collect_logs,
    help_category=HelpCategory.TROUBLESHOOT,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "--output",
                    short_name="-o",
                    help=messages.CLI_COLLECT_LOGS_OUTPUT,
                )
            ]
        )
    ],
)



import mock
import pytest

from uaclient.api.u.pro.attach.auto.configure_retry_service.v1 import (
    ConfigureRetryServiceOptions,
    _configure_retry_service,
)
from uaclient.daemon import retry_auto_attach
from uaclient.files import state_files

M_PATH = "uaclient.api.u.pro.attach.auto.configure_retry_service.v1."


@mock.patch(M_PATH + "system.create_file")
@mock.patch(M_PATH + "state_files.retry_auto_attach_options_file.write")
class TestConfigureRetryServiceV1:
    @pytest.mark.parametrize(
        "options, expected_options_write_calls, expected_create_file_calls",
        [
            (
                ConfigureRetryServiceOptions(),
                [mock.call(state_files.RetryAutoAttachOptions())],
                [mock.call(retry_auto_attach.FLAG_FILE_PATH)],
            ),
            (
                ConfigureRetryServiceOptions(enable=["cis"]),
                [
                    mock.call(
                        state_files.RetryAutoAttachOptions(enable=["cis"])
                    )
                ],
                [mock.call(retry_auto_attach.FLAG_FILE_PATH)],
            ),
            (
                ConfigureRetryServiceOptions(
                    enable=["cis"], enable_beta=["esm-infra"]
                ),
                [
                    mock.call(
                        state_files.RetryAutoAttachOptions(
                            enable=["cis"], enable_beta=["esm-infra"]
                        )
                    )
                ],
                [mock.call(retry_auto_attach.FLAG_FILE_PATH)],
            ),
        ],
    )
    def test_configure_retry_service(
        self,
        m_options_write,
        m_create_file,
        options,
        expected_options_write_calls,
        expected_create_file_calls,
        FakeConfig,
    ):
        _configure_retry_service(options, FakeConfig())
        assert expected_options_write_calls == m_options_write.call_args_list
        assert expected_create_file_calls == m_create_file.call_args_list



import abc
from typing import Any, Dict

from uaclient import config, contract


class AutoAttachInstance(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def is_viable(self) -> bool:
        """Return True if the machine is a viable AutoAttachCloudInstance."""
        pass

    @abc.abstractmethod
    def acquire_pro_token(self, cfg: config.UAConfig) -> str:
        """
        Cloud-specific implementation of acquiring the pro token using whatever
        method suits the platform
        """
        pass

    @abc.abstractmethod
    def should_poll_for_pro_license(self) -> bool:
        """
        Cloud-specific checks for whether the daemon should continously poll
        for Ubuntu Pro licenses.
        """
        pass

    @abc.abstractmethod
    def is_pro_license_present(self, *, wait_for_change: bool) -> bool:
        """
        Check for an Ubuntu Pro license
        """
        pass


class PublicCloudAutoAttachInstance(AutoAttachInstance, metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def identity_doc(self) -> Dict[str, Any]:
        """Return the identity document representing this cloud instance"""
        pass

    @property
    @abc.abstractmethod
    def cloud_type(self) -> str:
        """Return a string of the cloud type on which this instance runs"""
        pass

    def acquire_pro_token(self, cfg: config.UAConfig) -> str:
        """
        Cloud-specific implementation of acquiring the pro token using whatever
        method suits the platform
        """
        contract_client = contract.UAContractClient(cfg)
        token_response = contract_client.get_contract_token_for_cloud_instance(
            cloud_type=self.cloud_type, data=self.identity_doc
        )
        return token_response["contractToken"]
